"""
Unit and regression tests for the H5_Pdist class.
"""

# Import package, test suite, and other packages as needed
import wedap

import numpy as np
import pytest

# look at file coverage for testing
# pytest -v --cov=wedap
# produces .coverage binary file to be used by other tools to visualize 
# do not need 100% coverage, 80-90% is very high

# can have report in 
# $ pytest -v --cov=wedap --cov-report=html
# index.html to better visualize the test coverage

# decorator to skip in pytest
#@pytest.mark.skip

# TODO: test for trace, search_aux, skip_basis, get_total_data_array, get_all_weights
# maybe test more args like first_iter, last_iter, step_iter, H5save_out, data_proc, bins, histrange, p_units
# could also change to 1/2/3 dataset format

class Test_H5_Pdist():
    """
    Test each method of the H5_Pdist class.
    """
    h5 = "wedap/data/p53.h5"
    
    @pytest.mark.parametrize("Xname", ["pcoord", "dihedral_2"])
    def test_evolution(self, Xname):
        evolution = wedap.H5_Pdist(h5=self.h5, data_type="evolution", Xname=Xname)
        X, Y, Z = evolution.pdist()

        # X data is the variably filled array of instance pdist x values
        np.testing.assert_allclose(X, np.loadtxt(f"wedap/data/evolution_{Xname}_X.txt"))

        # Y data is just the WE iterations
        np.testing.assert_allclose(Y, 
            np.arange(evolution.first_iter, evolution.last_iter + 1, 1))

        # Z data is the pdist values of each iteration
        np.testing.assert_allclose(Z, np.loadtxt(f"wedap/data/evolution_{Xname}_Z.txt"))

    @pytest.mark.parametrize("Xname", ["pcoord", "dihedral_2"])
    def test_instant_1d(self, Xname):
        X, Y, Z = wedap.H5_Pdist(h5=self.h5, data_type="instant", Xname=Xname).pdist()
        np.testing.assert_allclose(X, np.loadtxt(f"wedap/data/instant_{Xname}_X.txt"))
        np.testing.assert_allclose(Y, np.loadtxt(f"wedap/data/instant_{Xname}_Y.txt"))
        
    @pytest.mark.parametrize("Xname", ["pcoord", "dihedral_2"])
    @pytest.mark.parametrize("Yname", ["dihedral_3", "dihedral_4"])
    def test_instant_2d(self, Xname, Yname):
        X, Y, Z = wedap.H5_Pdist(h5=self.h5, data_type="instant", Xname=Xname, Yname=Yname).pdist()
        np.testing.assert_allclose(X, 
            np.loadtxt(f"wedap/data/instant_{Xname}_{Yname}_X.txt"))
        np.testing.assert_allclose(Y, 
            np.loadtxt(f"wedap/data/instant_{Xname}_{Yname}_Y.txt"))
        np.testing.assert_allclose(Z, 
            np.loadtxt(f"wedap/data/instant_{Xname}_{Yname}_Z.txt"))
    
    # TODO along with average
    # def test_instant_3d(self):
    #     X, Y, Z = wedap.H5_Pdist(h5=self.h5, data_type="instant", Xname=Xname, Yname=Yname).pdist()
    #     np.testing.assert_allclose(X, 
    #         np.loadtxt(f"wedap/data/instant_{Xname}_{Yname}_X.txt"))
    #     np.testing.assert_allclose(Y, 
    #         np.loadtxt(f"wedap/data/instant_{Xname}_{Yname}_Y.txt"))
    #     np.testing.assert_allclose(Z, 
    #         np.loadtxt(f"wedap/data/instant_{Xname}_{Yname}_Z.txt"))

    @pytest.mark.parametrize("Xname", ["pcoord", "dihedral_2"])
    def test_average_1d(self, Xname):
        X, Y, Z = wedap.H5_Pdist(h5=self.h5, data_type="average", Xname=Xname).pdist()
        np.testing.assert_allclose(X, np.loadtxt(f"wedap/data/average_{Xname}_X.txt"))
        np.testing.assert_allclose(Y, np.loadtxt(f"wedap/data/average_{Xname}_Y.txt"))

    @pytest.mark.parametrize("Xname", ["pcoord", "dihedral_2"])
    @pytest.mark.parametrize("Yname", ["dihedral_3", "dihedral_4"])
    def test_average_2d(self, Xname, Yname):
        X, Y, Z = wedap.H5_Pdist(h5=self.h5, data_type="average", Xname=Xname, Yname=Yname).pdist()
        np.testing.assert_allclose(X, 
            np.loadtxt(f"wedap/data/average_{Xname}_{Yname}_X.txt"))
        np.testing.assert_allclose(Y, 
            np.loadtxt(f"wedap/data/average_{Xname}_{Yname}_Y.txt"))
        np.testing.assert_allclose(Z, 
            np.loadtxt(f"wedap/data/average_{Xname}_{Yname}_Z.txt"))



import wedap

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
import numpy as np

def plot_data_gen(h5, data_type, plot_mode, Xname, Yname=None, Zname=None, out=None):
    """
    Make plot and convert to npy binary data file.
    """
    plot = wedap.H5_Plot(h5=h5, data_type=data_type, plot_mode=plot_mode, 
                        Xname=Xname, Yname=Yname, Zname=Zname)
    plot.plot()
    fig = plot.fig

    #plt.show()
    #import sys; sys.exit(0)

    # If we haven't already shown or saved the plot, then we need to
    # draw the figure first...
    # tight since canvas is large
    fig.tight_layout(pad=0)
    fig.canvas.draw()

    # Now we can save it to a numpy array.
    data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    # save as np binary file
    if out:
        np.save(out, data)

h5 = "wedap/data/p53.h5"

# 1 dataset plots
for dt in ["evolution", "average", "instant"]:
    for pm in ["line"]:
        for x in ["pcoord", "dihedral_2"]:
            plot_data_gen(h5, dt, pm, x, out=f"wedap/data/plot_{dt}_{pm}_{x}.npy")

# 2 dataset plots
for dt in ["average", "instant"]:
    for pm in ["hist", "contour"]:
        for x, y in [["pcoord", "dihedral_2"], ["dihedral_2", "pcoord"]]:
            plot_data_gen(h5, dt, pm, x, y, out=f"wedap/data/plot_{dt}_{pm}_{x}_{y}.npy")

# 3 dataset plots
for dt in ["average", "instant"]:
    for pm in ["scatter3d"]:
        for x, y, z in [["pcoord", "dihedral_2", "dihedral_3"], 
                        ["dihedral_2", "pcoord", "dihedral_3"], 
                        ["dihedral_2", "dihedral_3", "pcoord"]]:
            plot_data_gen(h5, dt, pm, x, y, z, out=f"wedap/data/plot_{dt}_{pm}_{x}_{y}_{z}.npy")


"""
Functions for handling command-line input using argparse module.
"""

import argparse
import sys

# import and use gooey conditionally
# adapted from https://github.com/chriskiehl/Gooey/issues/296
try:
    import gooey
    #from gooey import Gooey
    #from gooey import GooeyParser
except ImportError:
    gooey = None

def flex_add_argument(f):
    """Make the add_argument accept (and ignore) the widget option."""

    def f_decorated(*args, **kwargs):
        kwargs.pop('widget', None)
        return f(*args, **kwargs)

    return f_decorated

# Monkey-patching a private class…
argparse._ActionsContainer.add_argument = \
    flex_add_argument(argparse.ArgumentParser.add_argument)

# Do not run GUI if it is not available or if command-line arguments are given.
if gooey is None or len(sys.argv) > 1:
    ArgumentParser = argparse.ArgumentParser

    def gui_decorator(f):
        return f
else:
    ArgumentParser = gooey.GooeyParser
    gui_decorator = gooey.Gooey(
        program_name='mdap',
        #navigation='TABBED',
        #advanced=True,
        suppress_gooey_flag=True,
        optional_cols=4, 
        default_size=(1000, 600),
        #tabbed_groups=True,
    )

# TODO: make tabs?
# then push new pip version
@gui_decorator
def create_cmd_arguments(): 
    """
    Use the `argparse` module to make the optional and required command-line
    arguments for `mdap`. 

    Parameters 
    ----------

    Returns
    -------
    argparse.ArgumentParser: 
        An ArgumentParser that is used to retrieve command line arguments. 
    """
    mdap_desc = "======================================================= \n" + \
                "=== molecular dynamics analysis and plotting (mdap) === \n" + \
                "======================================================= \n" + \
                "\nGiven an input (pre-calcualated) dataset from standard MD simulations, " + \
                "prepare probability distributions and plots." + \
                "Input data must be in >=2 column format: \n" + \
                "# use hashs at top of data file to indicate comments (skipped) \n" + \
                "COL1:Frame | COL2:Data | COL3:Data..."

    # create argument parser (gooey based if available)
    if gooey is None:
        parser = argparse.ArgumentParser(description=mdap_desc, 
                                        formatter_class=argparse.RawDescriptionHelpFormatter)
    else:
        parser = gooey.GooeyParser(description=mdap_desc, 
                                   formatter_class=argparse.RawDescriptionHelpFormatter)

    ##########################################################
    ############### REQUIRED ARGUMENTS #######################
    ##########################################################

    # create new group for required args 
#     required_args = parser.add_argument_group("Required Arguments") 

#     # create file flag  
#     required_args.add_argument("-h5", "--h5file", required = True, help = "The \
#         WESTPA west.h5 output file that will be analyzed.", action = "store", 
#         dest = "h5", type=str) 

    # specify the main group of args needed and shown on initial page
    # note that these are positional arguments to parser
    # so like $ plothist average
    # good for different tools, maybe good for movie
    #main = parser.add_subparsers(help="Main Arguments", dest="Main")
    # sub_main = main.add_parser('option1')
    # sub_main.add_argument("test")
    # sub_main2 = main.add_parser('option2')
    # sub_main2.add_argument("test")

    # test out gooey specific widgets
#     required = parser.add_argument_group("Required Arguments")
#     required.add_argument("-h5", "--h5file", #required=True, nargs="?",
#         default="west.h5", action="store", dest="h5", type=str,
#         help="The WESTPA west.h5 output file that will be analyzed. "
#              "Default 'west.h5'.", 
#         widget="FileChooser")

    ###########################################################
    ############### OPTIONAL ARGUMENTS ########################
    ###########################################################
    # nargs = '?' "One argument will be consumed from the command line if possible, 
        # and produced as a single item. If no command-line argument is present, 
        # the value from default will be produced."

    main = parser.add_argument_group("Main Arguments")
    optional = parser.add_argument_group("Optional Extra Arguments")

    main.add_argument("-dt", "--data-type", default="pdist", nargs="?",
                        dest="data_type", choices=("time", "pdist"),
                        help="Type of pdist dataset to generate, options are ",
                             # TODO
                        type=str) 
    main.add_argument("-pm", "--plot-mode", default="hist", nargs="?",
                        dest="plot_mode", choices=("hist", "hist_l", "contour", "contour_l", 
                                                   "contour_f", "bar", "line", "scatter3d"),
                        help="The type of plot desired.  "
                             "e.g. line for 1D, hist or contour for 2D and scatter3d for 3D.",
                        type=str)
    # TODO: allow a list of files, then multiple replicates can be addressed
    # * args is flexible number of values, which will be gathered into a list
    main.add_argument("-X", "-x", "--Xname", "--xname", default=None, nargs="*",
                        dest="Xname",
                        help="Target data name for x axis",
                        type=str)
    main.add_argument("-Y", "-y", "--Yname", "--yname", default=None, nargs="*",
                        dest="Yname",
                        help="Target data name for y axis.",
                        type=str)
    main.add_argument("-Z", "-z", "--Zname", "--zname", default=None, nargs="*",
                        dest="Zname", 
                        help="Target data name for z axis. Must use 'scatter3d' "
                        "for 'plot_mode'.",
                        type=str)
    # default to index 1 (2nd item of dataset)
    main.add_argument("-Xi", "-xi", "--Xindex", "--xindex", default=1, nargs="?", type=int,
                        dest="Xindex", help="Index in third dimension for >2D datasets.")
    main.add_argument("-Yi", "-yi", "--Yindex", "--yindex", default=1, nargs="?", type=int,
                        dest="Yindex", help="Index in third dimension for >2D datasets.")
    main.add_argument("-Zi", "-zi", "--Zindex", "--zindex", default=1, nargs="?", type=int,
                        dest="Zindex", help="Index in third dimension for >2D datasets.")
    # default to interval of 1 (process every frame)
    # TODO: update convention to be different from index
    main.add_argument("-Xint", "-xint", "--Xinterval", "--xinterval", default=1, nargs="?", type=int,
                        dest="Xinterval", help="Interval in third dimension for >2D datasets.")
    main.add_argument("-Yint", "-yint", "--Yinterval", "--yinterval", default=1, nargs="?", type=int,
                        dest="Yinterval", help="interval in third dimension for >2D datasets.")
    main.add_argument("-Zint", "-zint", "--Zinterval", "--zinterval", default=1, nargs="?", type=int,
                        dest="Zinterval", help="Interval in third dimension for >2D datasets.")
    main.add_argument("-hrx", "--histrange-x", default=None, nargs=2,
                      dest="histrange_x",
                      help="Ranges to consider for the x-axis, input "
                           "2 space-seperated floats : LB UB",
                      type=float)
    main.add_argument("-hry", "--histrange-y", default=None, nargs=2,
                      dest="histrange_y",
                      help="Ranges to consider for the y-axis, input "
                           "2 space-seperated floats : LB UB",
                      type=float)
    main.add_argument("-o", "--output", default=None,
                        dest="output_path",
                        help="The filename to which the plot will be saved. "
                             "Various image formats are available. You " 
                             "may choose one by specifying an extension. "
                             "\nLeave this empty if you don't want to save "
                             "the plot to a serperate file.",
                        type=str)
    # begin optional arg group
    # TODO: update to be first/last frame and add an interval arg
    optional.add_argument("-fi", "--first-iter", default=1, nargs="?",
                        dest="first_iter",
                        help="Plot data starting at iteration FIRST_ITER. "
                             "By default, plot data starting at the first "
                             "iteration in the specified west.h5 file.",
                        type=int)
    optional.add_argument("-li", "--last-iter", default=None, nargs="?",
                        dest="last_iter",
                        help="Plot data up to and including iteration LAST_ITER. "
                             "By default, plot data up to and including the last "
                             "iteration in the specified H5 file.",
                        type=int)
    # *: a flexible number of values, which will be gathered into a list
    # +: like *, but requiring at least one value
    optional.add_argument("--bins", default=[100, 100], nargs="+",
                        dest="bins",
                        help="Use BINS number of bins for histogramming X and Y. "
                             "Divide the range between the minimum and maximum "
                             "observed values into this many bins. Must input 1 bin "
                             "value per dimension (e.g. X and Y so '100 100')",
                        type=int)
    optional.add_argument("--pmin", default=None, nargs="?",
                        dest="p_min",
                        help="The minimum probability value limit. "
                             "This determines the cbar limits and contour levels.",
                        type=float)
    optional.add_argument("--pmax", default=None, nargs="?",
                        dest="p_max",
                        help="The maximum probability limit value. "
                             "This determines the cbar limits and contour levels.",
                        type=float)
    optional.add_argument("-pu", "--p-units", default="kT", nargs="?",
                        dest="p_units", choices=("kT", "kcal", "raw", "raw_norm"),
                        help="Can be 'kT' (default) or 'kcal'. "
                             "kT = -lnP, kcal/mol = -RT(lnP), "
                             "where RT=0.5922 at T(298K).",
                        type=str)
    optional.add_argument("-ts", "--timescale", default=10e6, nargs="?",
                        dest="timescale",
                        help="Default ps to µs (10e6). Converts frames to time.",
                        type=float)
    optional.add_argument("-ci", "--contour-interval", default=1, nargs="?",
                        dest="contour_interval",
                        help="If using plot-mode contour, "
                             "This sets the interval of contour level.",
                        type=float)
    optional.add_argument("-cl", "--contour-levels", default=None, nargs="*",
                        dest="contour_levels",
                        help="If using plot-mode contour, "
                             "This overrides and sets the contour levels manually.",
                        type=float)
    optional.add_argument("-sl", "--smoothing-level", default=None, 
                        dest="smoothing_level",
                        help="Smooth data (plotted as histogram or contour"
                             " levels) using a gaussian filter with sigma="
                             "SMOOTHING_LEVEL.",
                        type=float)
    optional.add_argument("-T", "--temp", default=298, nargs="?",
                        dest="T", help="Used with kcal/mol 'p-units'.",
                        type=int)
    optional.add_argument("-jp", "--joint-plot", default=False,
                          dest="jointplot",
                          help="Optionally include marginal plots to create "
                               "a joint plot from 2D pdist.",
                          action="store_true")
    optional.add_argument("-3d", "--proj3d", default=False,
                          dest="proj3d",
                          help="Make a 3d projection plot",
                          action="store_true")
    optional.add_argument("--style", default="default", nargs="?",
                        dest="style",
                        help="mpl style, can use default, None, or custom. "
                             "Edit the wedap/styles/default.mplstyle file to "
                             "change default wedap plotting style options.",
                        type=str)
    # TODO: prob cant use custom outside of list
    optional.add_argument("--cmap", default="viridis", nargs="?",
                        dest="cmap",
                        help="mpl colormap name.",
                        type=str)
    optional.add_argument("--color",
                        dest="color", help="Color for 1D plots and trace plots.",
                        widget="ColourChooser")

    # create optional flag to not output plot to console screen
    optional.add_argument("-nots", "--no-output-to-screen",
                        dest = "no_output_to_screen",
                        help = "Include this argument to not output the plot to "
                        "your display.", 
                        action= "store_true") 
    optional.add_argument("-npb", "--no-progress-bar",
                        dest = "no_pbar",
                        help = "Include this argument to not output the tqdm progress bar.",
                        action= "store_true")

    # plot tracing arg group (TODO: include this as trace-frame and can also trace-val)
    trace = parser.add_argument_group("Optional Plot Tracing", 
                                       description="Plot a trace on top of the pdist.")
    trace_group = trace.add_mutually_exclusive_group()
    # type to float for val inside tuple, 
    # and nargs to 2 since it is interpreted as a 2 item tuple or list
    trace_group.add_argument("--trace-frame", default=None, nargs=1,
                             dest="trace_frame",
                             help="Trace and plot a single continuous trajectory based "
                                 "off of 1 ints to specify frame number.",
                             type=int)
    trace_group.add_argument("--trace-val", default=None, nargs=2,
                             dest="trace_val",
                             help="Trace and plot the trajectory up this value based "
                                  "off of 2 space-seprated floats : Xvalue Yvalue",
                             type=float)

    ##########################################################
    ############### FORMATTING ARGUMENTS #####################
    ##########################################################

    formatting = parser.add_argument_group("Plot Formatting Arguments") 

    formatting.add_argument("--xlabel", dest="xlabel", type=str)
    formatting.add_argument("--xlim", help="LB UB", dest="xlim", nargs=2, type=float)
    formatting.add_argument("--ylabel", dest="ylabel", type=str)
    formatting.add_argument("--ylim", help="LB UB", dest="ylim", nargs=2, type=float)
    formatting.add_argument("--title", dest="title", type=str)
    formatting.add_argument("--suptitle", dest="suptitle", type=str)
    formatting.add_argument("--cbar-label", dest="cbar_label", type=str)
    formatting.add_argument("--grid", dest="grid", default=False, action="store_true")

    # return the argument parser
    return parser 


# TODO: adjust all to fix str/int/type auto
def handle_command_line(argument_parser): 
    """
    Take command line arguments, check for issues, return the arguments. 

    Parameters
    ----------
    argument_parser : argparse.ArgumentParser 
        The argument parser that is returned in `create_cmd_arguments()`.
    
    Returns
    -------
    argparse.NameSpace
        Contains all arguments passed into wedap.
    
    Raises
    ------  
    Prints specific issues to terminal.
    """
    # retrieve args
    args = argument_parser.parse_args() 

    # h5 file and file exists
    # if not os.path.exists(args.file) or not ".h5" in args.file:  
    #     # print error message and exits
    #     sys.exit("Must input file that exists and is in .h5 file format.")

    # if not args.percentage.isdigit():  # not correct input   
    #     # print out any possible issues
    #     print("You must input a percentage digit. EXAMPLES: \
    #     \n '-p 40' \n You CANNOT add percent sign (eg. 50%) \n You \
    #     CANNOT add decimals (eg. 13.23)") 
    #     sys.exit(0) # exit program

    # # ignore if NoneType since user doesn't want --maxEnsembleSize parameter
    # if args.max_ensemble_size is None: 
    #     pass

    # elif not args.max_ensemble_size.isdigit(): # incorrect input 
    #     # needs to be whole number
    #     print("You must input a whole number with no special characters (eg. 4).")  
    #     sys.exit(0) # exit program 

    # elif args.max_ensemble_size is '0': # user gives 0 to --maxEnsembleSize flag 
    #     print("You cannot input '0' to --maxEnsembleSize flag.")
    #     sys.exit(0) # exit program 

    return args


"""
Convert MD analyzed data to pdists.

TODO:
    Maybe could include an arg for custom weights
"""

import sys
import numpy as np
from wedap import H5_Pdist

# Suppress divide-by-zero in log
np.seterr(divide='ignore', invalid='ignore')

class MD_Pdist(H5_Pdist):
    """
    These class methods generate probability distributions from input MD data files.
    """
    # TODO: is setting aux_y to None the best approach to 1D plot settings?
    # TODO: add step-iter
    def __init__(self, data_type=None, Xname=None, Xindex=1, Yname=None, Yindex=1, 
                 Zname=None, Zindex=1, Xinterval=1, Yinterval=1, Zinterval=1, data_proc=None, 
                 first_iter=1, last_iter=None, bins=(100,100), p_units='kT', T=298, 
                 histrange_x=None, histrange_y=None, no_pbar=False, timescale=10**6,
                 *args, **kwargs):
        """
        TODO: add XYZ interval to proc (default 1)

        Parameters
        ----------
        data_type : str
            'timeseries' for 1 dataset, or 'pdist' for 2D. scatter for 3d? (TODO)
        Xname : str or list of str
            target data for x axis, default None.
        Xindex : int
            If X.ndim > 2, use this to index.
        Yname : str or list of str
            target data for y axis, default None.
        Yindex : int
            If Y.ndim > 2, use this to index.
        Zname : str or list of str
            target data for z axis, default None. 
            Use this if you want to use a dataset instead of pdist for Z axis.
            This will be best plotted as a scatter plot with Z as the marker color.
            Instead of returning the pdist, only the XYZ datasets will be returned.
            This is becasue the weights/pdist isn't considered.
        Zindex : int
            If Z.ndim > 2, use this to index.
        Xinterval, Yinterval, Zinterval : int
            Interval for processing dataset. E.g. 10 = every 10 frames.
        data_proc : function or tuple of functions
            Of the form f(data) where data has rows=segments, columns=frames until tau, depth=data dims.
            The input function must return a processed array of the same shape and formatting.
        first_iter : int
            Default start plot at iteration 1 data.
        last_iter : int
            Last iteration data to include, default is the last recorded iteration in the west.h5 file. 
            Note that `instant` type pdists only depend on last_iter.
        bins : tuple of ints (TODO: maybe the tuple isn't user friendly for 1 dim?)
            Histogram bins in pdist data to be generated for x and y datasets, default both 100.
        p_units : str
            Can be 'kT' (default), 'kcal', 'raw', or 'raw_norm'.
            kT = -lnP, kcal/mol = -RT(lnP), where RT = 0.5922 at `T` Kelvin.
            'raw' is the raw probabilities and 'raw_norm' is the raw probabilities P(max) normalized.
        T : int
            Temperature if using kcal/mol.
        histrange_x, histrange_y : list or tuple of 2 floats or ints
            Optionally put custom bin ranges.
        no_pbar : bool
            Optionally do not include the progress bar for pdist generation.
        timescale : int
            Default ps to µs (10**6). Converts frames to time.
        TODO: maybe also binsfromexpression?
        """
        if data_type is None:
            raise ValueError("Must input valid data_type: `timeseries`, `pdist`, or ")
        else:
            self.data_type = data_type
        self.p_units = p_units
        self.T = T

        self.Xname = Xname
        self.Xindex = Xindex
        self.Yname = Yname
        self.Yindex = Yindex
        self.Zname = Zname
        self.Zindex = Zindex
        
        self.Xinterval = Xinterval
        self.Yinterval = Yinterval
        self.Zinterval = Zinterval

        # raw data processing function
        # TODO: allow for 1-3 functions as tuple input, right now one function only
        self.data_proc = data_proc

        # default to last
        if last_iter is not None:
            self.last_iter = last_iter
        elif last_iter is None:
            pass # entire trajectory
        self.first_iter = first_iter

        self.bins = bins
        self.histrange_x = histrange_x
        self.histrange_y = histrange_y
        self.no_pbar = no_pbar
        self.timescale = timescale

    def _get_md_data(self, names, index, interval):
        """
        Return MD data in array.
        """
        # grab dataset (make sure 2D for proper indexing)
        # TODO: switch to pre-cast array?
        data = []
        # handle multiple file name list
        # TODO: handle direct ndarray inputs as well as file str
        for name in names:
            # for .npy binary files or pkl files
            if name[-4:] in [".npy", ".npz", ".pkl"]:
                data_item = np.load(name, allow_pickle=True)
            else:
                # TODO: note that genfromtxt can handle multiple items in list, could this help?
                data_item = np.genfromtxt(name)
            # for 1D datasets, need to standardize to 2D, but as a new column
            if data_item.ndim < 2:
                data_item = data_item[:, np.newaxis]
            data.append(data_item[::interval, index])
        # combine into a single array
        data = np.concatenate(data)

        return data

    def timeseries(self):
        """
        Returns
        -------
        X : ndarray
        Y : ndarray
        """
        # could get time from frame column
        #time = np.concatenate([np.genfromtxt(i)[::self.Xinterval, 0] for i in self.Xname])

        X = self._get_md_data(self.Xname, self.Xindex, self.Xinterval)

        # or can just get it from n rows
        time = np.arange(0, X.shape[0], self.Xinterval)
        time = np.divide(time, self.timescale)

        return time, X

    def pdist_1d(self):
        """
        Returns
        -------
        X : ndarray
        Y : ndarray
        """
        #X = np.concatenate([np.genfromtxt(i)[::self.Xinterval, self.Xindex] for i in self.Xname])
        X = self._get_md_data(self.Xname, self.Xindex, self.Xinterval)
    
        # get rid of nan values: return array without (not) True nan values
        #X = X[np.logical_not(np.isnan(X))]

        # numpy equivalent to: ax.hist2d(c2[:,1], aux)
        hist, x_edges = np.histogram(X, bins=self.bins[0], range=self.histrange_x)
        # let each row list bins with common y range
        hist = np.transpose(hist)
        # convert histogram counts to p_units
        hist = self._normalize(hist)
        # get bin midpoints
        midpoints_x = (x_edges[:-1] + x_edges[1:]) / 2

        return midpoints_x, hist

    def pdist_2d(self):
        """
        Returns
        -------
        X : ndarray
        Y : ndarray
        Z : ndarray
        """
        X = self._get_md_data(self.Xname, self.Xindex, self.Xinterval)
        Y = self._get_md_data(self.Yname, self.Yindex, self.Yinterval)
    
        try:
            # numpy equivalent to: ax.hist2d(c2[:,1], aux)
            hist, x_edges, y_edges = np.histogram2d(X, Y, bins=self.bins, 
                                                    range=[self.histrange_x, self.histrange_y])
        except ValueError as e:
            print(f"{e}: do the size of the x {X.shape} and y {Y.shape} inputs match? --Xinterval and --Yinterval can be adjusted.")
            sys.exit(0)
        # let each row list bins with common y range
        hist = np.transpose(hist)
        # convert histogram counts to p_units
        hist = self._normalize(hist)
        # get bin midpoints
        midpoints_x = (x_edges[:-1] + x_edges[1:]) / 2
        midpoints_y = (y_edges[:-1] + y_edges[1:]) / 2

        return midpoints_x, midpoints_y, hist
        
    def pdist_3d(self):
        """
        Returns
        -------
        X : ndarray
        Y : ndarray
        Z : ndarray
        """
        X = self._get_md_data(self.Xname, self.Xindex, self.Xinterval)
        Y = self._get_md_data(self.Yname, self.Yindex, self.Yinterval)
        Z = self._get_md_data(self.Zname, self.Zindex, self.Zinterval)

        return X, Y, Z

    def pdist(self):
        """
        Main public method with pdist generation controls.
        """ 
        # TODO: need to consolidate the Y 2d vs 1d stuff somehow

        # TODO: if I can get rid of this or optimize it, I can then use the 
            # original methods of each pdist by themselves
        # TODO: only if histrange is None
        # if self.histrange_x is None:
        #     # get the optimal histrange
        #     self.histrange_x = self._get_histrange(self.Xname, self.Xindex)
        # # if using 2D pdist
        # # TODO: needs to handle array input or None input
        # if isinstance(self.Yname, (str, np.ndarray)) and self.histrange_y is None:
        #     self.histrange_y = self._get_histrange(self.Yname, self.Yindex)

        # return timeseries data
        if self.data_type == "time":
            X, Y = self.timeseries()
            return X, Y, np.ones((X.shape[0]))
        # return pdist data
        # TODO: t/e block to catch mismatched data lengths and print lengths
        elif self.data_type == "pdist":
            # for 3D datasets, e.g. scatter
            if self.Yname and self.Zname:
                return self.pdist_3d()
            # if 2D, return 2D dataset
            if self.Yname:
                return self.pdist_2d()
            # otherwise return 1D
            else:
                X, Y = self.pdist_1d()
                # TODO: fake 3D output for now to accomodate main.py
                return X, Y, np.ones(X.shape[0])



"""
Main call.
"""
from wedap.h5_pdist import *
from wedap.h5_plot import *
from wedap.command_line import *

#from wedap.search_aux import *
#from wedap.h5_plot_trace import *

# TODO: change to logging style instead of stdout
#import logging

# for accessing package data: mpl styles
import pkgutil 
import os

def main():

    """
    Command line
    """
    # Create command line arguments with argparse
    argument_parser = create_cmd_arguments()
    # Retrieve list of args
    args = handle_command_line(argument_parser)

    """
    Generate pdist and plot it
    """
    if args.style == "default":
        # get the style parameters from package data
        # currently writes into temp file, could be more efficient (TODO)
        style = pkgutil.get_data(__name__, "styles/default.mplstyle")
        # pkgutil returns binary string, decode it first and make temp file
        with open("style.temp", "w+") as f:
            f.write(style.decode())
        plt.style.use("style.temp")
        # clean up temp style file
        os.remove("style.temp")
    elif args.style != "default" and args.style != "None":
        plt.style.use(args.style)

    # if args.p_units == "kT":
    #     cbar_label = "$-\ln\,P(x)$"
    # elif args.p_units == "kcal":
    #     cbar_label = "$-RT\ \ln\, P\ (kcal\ mol^{-1})$"
    # elif args.p_units == "raw":
    #     cbar_label = "Counts"
    # elif args.p_units == "raw_norm":
    #     cbar_label = "Normalized Counts"

    # a poor workaround for now for the weighted arg
    # this is only to make the gooey formatting look nicer in terms of the checkbox
    if args.not_weighted is True:
        args.weighted = False
    elif args.not_weighted is False:
        args.weighted = True

    # accounting for joint plot option (TODO: better way?)
    # if args.jointplot:
    #     temp_p_units = "raw"
    # else:
    #     temp_p_units = args.p_units

    # always output XYZ with fake Z for 1D, makes this part easier/less verbose
    #pdist = H5_Pdist()
    # plot = H5_Plot(# H5_Pdist args
    #                 h5=args.h5, data_type=args.data_type, Xname=args.Xname,
    #                 Xindex=args.Xindex, Yname=args.Yname, Yindex=args.Yindex, Zname=args.Zname, 
    #                 Zindex=args.Zindex, first_iter=args.first_iter, skip_basis=args.skip_basis,
    #                 last_iter=args.last_iter, bins=args.bins, T=args.T,
    #                 weighted=weighted, p_units=args.p_units, no_pbar=args.no_pbar,
    #                 histrange_x=args.histrange_x, histrange_y=args.histrange_y,
    #                 # H5_Plot args
    #                 plot_mode=args.plot_mode, cmap=args.cmap,
    #                 contour_interval=args.contour_interval, p_min=args.p_min,
    #                 p_max=args.p_max, cbar_label=cbar_label, color=args.color,
    #                 smoothing_level=args.smoothing_level, jointplot=args.jointplot,
    #                 **args_dict)
    
    # vars converts from Namespace object to dict
    plot = H5_Plot(**vars(args))

    # had to adjust this since joint_plots require raw dist, so coupled pdist/plot needed
    # X, Y, Z = pdist.pdist()
    # plot = H5_Plot(X, Y, Z, plot_mode=args.plot_mode, cmap=args.cmap,
    #                contour_interval=args.contour_interval, p_min=args.p_min,
    #                p_max=args.p_max, cbar_label=cbar_label, color=args.color,
    #                smoothing_level=args.smoothing_level, jointplot=args.jointplot)
    # 2D plot with cbar
    # TODO: can this be done better?
    if args.Yname or args.data_type == "evolution":
        try:
            plot.plot(cbar=True)
        except AttributeError as e:
            print(f"{e}: Attempting to plot an {args.data_type} dataset using ")
            print(f"a {args.plot_mode} type plot. Is this what you meant to do?")
            sys.exit(0)
    # 1D plot that isn't evolution
    elif args.Yname is None and args.Zname is None and args.data_type != "evolution":
        plot.plot(cbar=False)

    """
    Trace (Optional Argument)
    """
    # default to white if no color provided
    if args.color is None:
        args.color = "white"
    if args.trace_seg is not None:
        plot.plot_trace(args.trace_seg, color=args.color, ax=plot.ax)
    if args.trace_val is not None:
        iter, seg = plot.search_aux_xy_nn(args.trace_val[0], args.trace_val[1])
        plot.plot_trace((iter,seg), color=args.color, ax=plot.ax)

    """
    Plot formatting
    """
    # if no xlabel is given, create default label
    if args.xlabel is None:
        plot.ax.set_xlabel(args.Xname + " i" + str(plot.Xindex))

    # if no ylabel is given, create default label of Yname or "WE Iteration"
    if args.ylabel is None:
        if args.data_type == "evolution":
            plot.ax.set_ylabel("WE Iteration")
        if args.Yname:
            plot.ax.set_ylabel(args.Yname + " i" + str(plot.Yindex))

    # if cbar_label is given set as cbar_label
    if args.cbar_label:
        plot.cbar.set_label(args.cbar_label, labelpad=14)
    # if using scatter3d and no label is given, create default label
    elif args.plot_mode == "scatter3d":
        # for 3d projected
        if args.proj3d is True:
            plot.ax.set_zlabel(args.Zname + " i" + str(plot.Zindex))
        # for 2d scatter with cbar
        else:
            plot.cbar.set_label(args.Zname + " i" + str(plot.Zindex))
    

    """
    Show and/or save the final plot
    """
    # fig vs plt shouldn't matter here (needed to go plt for mosaic)
    #plot.fig.tight_layout()
    plt.tight_layout()
    if args.output_path is not None:
        # fig vs plt shouldn't matter here (needed to go plt for mosaic)
        #plot.fig.savefig(args.output_path)
        plt.savefig(args.output_path)
        #logging.info(f"Plot was saved to {args.output_path}")
        print(f"Plot was saved to {args.output_path}")
    if args.no_output_to_screen:
        pass
    else:
        plt.show()
        #plot.fig.show() # only for after event loop starts e.g. with plt.show()

# if python file is being used 
if __name__ == "__main__": 
    main()


"""
Main call.
"""
import matplotlib.pyplot as plt

from mdap.md_pdist import *
from mdap.md_plot import *
from mdap.command_line import *

# TODO: change to logging style instead of stdout
#import logging

# for accessing package data: mpl styles
import pkgutil 
import os

def main():

    """
    Command line
    """
    # Create command line arguments with argparse
    argument_parser = create_cmd_arguments()
    # Retrieve list of args
    args = handle_command_line(argument_parser)

    """
    Generate pdist and plot it
    """
    if args.style == "default":
        # get the style parameters from package data
        # currently writes into temp file, could be more efficient (TODO)
        style = pkgutil.get_data(__name__, "styles/default.mplstyle")
        # pkgutil returns binary string, decode it first and make temp file
        with open("style.temp", "w+") as f:
            f.write(style.decode())
        plt.style.use("style.temp")
        # clean up temp style file
        os.remove("style.temp")
    elif args.style != "default" and args.style != "None":
        plt.style.use(args.style)

    # # for jointplots, save original p_units and run pdist with raw
    # og_p_units = args.p_units
    # #print(args.jointplot)
    # if args.jointplot is True:
    #     #vars(args)["p_units"] = "raw"
    #     args.p_units = "raw"

    # pdist = MD_Pdist(**vars(args))
    # X, Y, Z = pdist.pdist()

    # # TODO: update to MD_Plot for more flexibility and customization
    # # plot = MD_Plot(X, Y, Z, plot_mode=args.plot_mode, cmap=args.cmap,
    # #                contour_interval=args.contour_interval, p_min=args.p_min,
    # #                p_max=args.p_max, cbar_label=cbar_label, color=args.color,
    # #                smoothing_level=args.smoothing_level, jointplot=args.jointplot)
    # # plot = plot.plot()

    # # TODO: note that jointplot will not work well since p_units are not available
    # # for jointplot, use og p_units to convert from raw
    # if args.jointplot:
    #     args.p_units = og_p_units
    # plot = H5_Plot(X, Y, Z, **vars(args))
    # plot.plot()

    # MD_Plot use with implicit use of MD_Pdist
    plot = MD_Plot(**vars(args))
    plot.plot()

    # attempt to split to seperate pdist and plot classes
    # X, Y, Z  = MD_Pdist(**vars(args)).pdist()
    # plot = MD_Plot(**vars(args))
    # plot.plot()

    """
    Trace (Optional Argument) TODO: adapt to stdMD
    """
    # default to white if no color provided
    if args.color is None:
        args.color = "white"
    if args.trace_frame is not None: # TODO
        plot.plot_trace(args.trace_frame, color=args.color, ax=plot.ax)
    if args.trace_val is not None:
        iter, seg = plot.search_aux_xy_nn(args.trace_val[0], args.trace_val[1])
        plot.plot_trace((iter,seg), color=args.color, ax=plot.ax)

    """
    Plot formatting
    """
    # if cbar_label is given set as cbar_label, otherwise try to find a good label
    if args.cbar_label:
        cbar_label = args.cbar_label
    elif args.p_units == "kT":
        cbar_label = "$-\ln\,P(x)$"
    elif args.p_units == "kcal":
        cbar_label = "$-RT\ \ln\, P\ (kcal\ mol^{-1})$"
    elif args.p_units == "raw":
        cbar_label = "Counts"
    elif args.p_units == "raw_norm":
        cbar_label = "Normalized Counts"
    # if using scatter3d and no label is given, create default label
    if args.plot_mode == "scatter3d" and args.cbar_label is None:
        cbar_label = plot.Zname[0] + " i" + str(plot.Zindex)
    # if there is a cbar object set label
    if hasattr(plot, "cbar"):
        plot.cbar.set_label(cbar_label, labelpad=14)
    # if using proj3d, add z axis label
    if args.proj3d is True:
        plot.ax.set_zlabel(cbar_label)

    # if no label is given, create default label (default to first item in XYZname list)
    if args.xlabel is None:
        # unless timeseries, then use time label
        if args.data_type == "time":
            plot.ax.set_xlabel(f"Time (frames x {str(args.timescale)})")
        else:
            plot.ax.set_xlabel(plot.Xname[0] + " i" + str(plot.Xindex))
    if args.ylabel is None:
        # use Xname on Y if timeseries, otherwise use Yname on Y
        if args.data_type == "time":
            plot.ax.set_ylabel(plot.Xname[0] + " i" + str(plot.Xindex))
        # if Y data is given
        elif args.Yname:
            plot.ax.set_ylabel(plot.Yname[0] + " i" + str(plot.Yindex))
        # otherwise it will be something like 1d pdist
        else:
            plot.ax.set_ylabel(cbar_label)

    """
    Show and/or save the final plot
    """
    # fig vs plt shouldn't matter here (needed to go plt for mosaic)
    #plot.fig.tight_layout()
    plt.tight_layout()
    if args.output_path is not None:
        # fig vs plt shouldn't matter here (needed to go plt for mosaic)
        #plot.fig.savefig(args.output_path)
        plt.savefig(args.output_path)
        #logging.info(f"Plot was saved to {args.output_path}")
        print(f"Plot was saved to {args.output_path}")
    if args.no_output_to_screen:
        pass
    else:
        plt.show()
        #plot.fig.show() # only for after event loop starts e.g. with plt.show()

# if python file is being used 
if __name__ == "__main__": 
    main()


"""
Functions for handling command-line input using argparse module.

TODO: add option to split pdist and plot once the pdist to txt feature is done
      so the pdist txt file could be: X_column, Y_column, Z_matrix_columns.
      This would be something like pdist.dap default and option to check and 
      read in a wedap or westpa pdist file for pdist function.
"""

import argparse
import sys

# import and use gooey conditionally
# adapted from https://github.com/chriskiehl/Gooey/issues/296
try:
    import gooey
    #from gooey import Gooey
    #from gooey import GooeyParser
except ImportError:
    gooey = None

def flex_add_argument(f):
    """Make the add_argument accept (and ignore) the widget option."""

    def f_decorated(*args, **kwargs):
        kwargs.pop('widget', None)
        return f(*args, **kwargs)

    return f_decorated

# Monkey-patching a private class…
argparse._ActionsContainer.add_argument = \
    flex_add_argument(argparse.ArgumentParser.add_argument)

# Do not run GUI if it is not available or if command-line arguments are given.
if gooey is None or len(sys.argv) > 1:
    ArgumentParser = argparse.ArgumentParser

    def gui_decorator(f):
        return f
else:
    ArgumentParser = gooey.GooeyParser
    gui_decorator = gooey.Gooey(
        program_name='wedap',
        #navigation='TABBED',
        #advanced=True,
        suppress_gooey_flag=True,
        optional_cols=4, 
        default_size=(1000, 600),
        #tabbed_groups=True,
    )

# TODO: make tabs?
# then push new pip version
@gui_decorator
def create_cmd_arguments(): 
    """
    Use the `argparse` module to make the optional and required command-line
    arguments for the `wedap`. 

    Parameters 
    ----------

    Returns
    -------
    argparse.ArgumentParser: 
        An ArgumentParser that is used to retrieve command line arguments. 
    """
    wedap_desc = "============================================================ \n" + \
                 "=== weighted ensemble data analysis and plotting (wedap) === \n" + \
                 "============================================================ \n" + \
                 "\nGiven an input west.h5 file from a successful WESTPA simulation, " + \
                 "prepare probability distributions and plots." + \
                 "\nSee the documentation for usage and examples: https://darianyang.github.io/wedap" + \
                 "\n\n" + \
                 "wedap can be used with 3 different --data-type (-dt) args: " + \
                 "\n\t`evolution` (default), `average`, and `instant`" + \
                 "\n\nAvailable --plot-mode (-pm) options are: " + \
                 "\n\t1D: `line`" + \
                 "\n\t2D: `hist` (default), `hist_l` (hist with contour lines), " + \
                 "\n\t    `contour` (lines and fill), `contour_l` (lines only), `contour_f` (fill only)" + \
                 "\n\t3D: `scatter3d`" + \
                 "\n\nExamples\n--------" + \
                 "\nEvolution plot of your progress coordinate:" + \
                 "\n\t$ wedap -W west.h5 (default) -dt evolution (default) -X pcoord (default)" + \
                 "\n\n1D average probability distribution of your progress coordinate:" + \
                 "\n\t$ wedap -W west.h5 -dt average -pm line" + \
                 "\n\n2D average probability distribution of pcoord 1 and 2:" + \
                 "\n\t$ wedap -W west.h5 -dt average -X pcoord (default) -Xi 0 (default) -Y pcoord -Yi 1" + \
                 "\n\n1D instant probability distribution of an aux dataset:" + \
                 "\n\t$ wedap -W west.h5 -dt instant -pm line -X auxname" + \
                 "\n\n3D scatter of your progress coordinates and aux data" + \
                 "\n\t$ wedap -W west.h5 -dt average -pm scatter3d -X pcoord -Xi 0 -Y pcoord -Yi 1 -Z auxname -Zi 0" + \
                 "\n\n2D average contour plot of 2 aux datasets for iterations 100 to 200 with probability limits in kcal/mol." + \
                 "\n\t$ wedap -dt average -pm contour -X auxname -Y auxname -fi 100 -li 200 --pmin 0 --pmax 20 --p-units kcal"

    # create argument parser (gooey based if available)
    if gooey is None:
        parser = argparse.ArgumentParser(description=wedap_desc, 
                                        formatter_class=argparse.RawDescriptionHelpFormatter)
    else:
        parser = gooey.GooeyParser(description=wedap_desc, 
                                   formatter_class=argparse.RawDescriptionHelpFormatter)

    ##########################################################
    ############### REQUIRED ARGUMENTS #######################
    ##########################################################

    # create new group for required args 
#     required_args = parser.add_argument_group("Required Arguments") 

#     # create file flag  
#     required_args.add_argument("-h5", "--h5file", required = True, help = "The \
#         WESTPA west.h5 output file that will be analyzed.", action = "store", 
#         dest = "h5", type=str) 

    # specify the main group of args needed and shown on initial page
    # note that these are positional arguments to parser
    # so like $ plothist average
    # good for different tools, maybe good for movie
    #main = parser.add_subparsers(help="Main Arguments", dest="Main")
    # sub_main = main.add_parser('option1')
    # sub_main.add_argument("test")
    # sub_main2 = main.add_parser('option2')
    # sub_main2.add_argument("test")

    # test out gooey specific widgets
    required = parser.add_argument_group("Required Arguments")
    required.add_argument("-W", "-h5", "--h5file", #required=True, nargs="?",
        default="west.h5", action="store", dest="h5", type=str,
        help="The WESTPA west.h5 output file that will be analyzed. "
             "Default 'west.h5'.", 
        widget="FileChooser")

    ###########################################################
    ############### OPTIONAL ARGUMENTS ########################
    ###########################################################
    # nargs = '?' "One argument will be consumed from the command line if possible, 
        # and produced as a single item. If no command-line argument is present, 
        # the value from default will be produced."

    main = parser.add_argument_group("Main Arguments")
    optional = parser.add_argument_group("Optional Extra Arguments")

    main.add_argument("-dt", "--data-type", default="evolution", nargs="?",
                        dest="data_type", choices=("evolution", "average", "instant"),
                        help="Type of pdist dataset to generate, options are "
                             "'evolution' (1 dataset); " 
                             "'average' or 'instance' (1 or 2 or 3 datasets).",
                        type=str) 
    main.add_argument("-pm", "--plot-mode", default="hist", nargs="?",
                        dest="plot_mode", choices=("hist", "hist_l", "contour", "contour_l", 
                                                   "contour_f", "bar", "line", "scatter3d"),
                        help="The type of plot desired.  "
                             "e.g. line for 1D, hist or contour for 2D and scatter3d for 3D.",
                        type=str)
    main.add_argument("-X", "-x", "--Xname", "--xname", default="pcoord", nargs="?",
                        dest="Xname", 
                        help="Target data name for x axis. Default 'pcoord', "
                        "can also be any aux dataset name in your h5 file.",
                        type=str)
    main.add_argument("-Y", "-y", "--Yname", "--yname", default=None, nargs="?",
                        dest="Yname", 
                        help="Target data name for y axis. Default 'None', "
                        "can be 'pcoord' or any aux dataset name in your h5 file.",
                        type=str)
    main.add_argument("-Z", "-z", "--Zname", "--zname", default=None, nargs="?",
                        dest="Zname", 
                        help="Target data name for z axis. Must use 'scatter3d' "
                        "for 'plot_mode'. Can be 'pcoord' or any aux dataset name "
                        "in your h5 file.",
                        type=str)
    main.add_argument("-Xi", "-xi", "--Xindex", "--xindex", default=0, nargs="?", type=int,
                        dest="Xindex", help="Index in third dimension for >2D datasets.")
    main.add_argument("-Yi", "-yi", "--Yindex", "--yindex", default=0, nargs="?", type=int,
                        dest="Yindex", help="Index in third dimension for >2D datasets.")
    main.add_argument("-Zi", "-zi", "--Zindex", "--zindex", default=0, nargs="?", type=int,
                        dest="Zindex", help="Index in third dimension for >2D datasets.")
    main.add_argument("-o", "--output", default=None,
                        dest="output_path",
                        help="The filename to which the plot will be saved. "
                             "Various image formats are available. You " 
                             "may choose one by specifying an extension. "
                             "\nLeave this empty if you don't want to save "
                             "the plot to a serperate file.",
                        type=str)
    # begin optional arg group
    optional.add_argument("-fi", "--first-iter", default=1, nargs="?",
                        dest="first_iter",
                        help="Plot data starting at iteration FIRST_ITER. "
                             "By default, plot data starting at the first "
                             "iteration in the specified west.h5 file.",
                        type=int)
    optional.add_argument("-li", "--last-iter", default=None, nargs="?",
                        dest="last_iter",
                        help="Plot data up to and including iteration LAST_ITER. "
                             "By default, plot data up to and including the last "
                             "iteration in the specified H5 file.",
                        type=int)
    # *: a flexible number of values, which will be gathered into a list
    # +: like *, but requiring at least one value
    optional.add_argument("--bins", default=[100, 100], nargs="+",
                        dest="bins",
                        help="Use BINS number of bins for histogramming X and Y. "
                             "Divide the range between the minimum and maximum "
                             "observed values into this many bins. Must input 1 bin "
                             "value per dimension (e.g. X and Y so '100 100')",
                        type=int)
    optional.add_argument("-hrx", "--histrange-x", default=None, nargs=2,
                          dest="histrange_x",
                          help="Ranges to consider for the x-axis, input "
                               "2 space-seperated floats : LB UB",
                          type=float)
    optional.add_argument("-hry", "--histrange-y", default=None, nargs=2,
                          dest="histrange_y",
                          help="Ranges to consider for the y-axis, input "
                               "2 space-seperated floats : LB UB",
                          type=float)
    optional.add_argument("--pmin", "--p-min", default=None, nargs="?",
                        dest="p_min",
                        help="The minimum probability value limit. "
                             "This determines the cbar limits and contour levels.",
                        type=float)
    optional.add_argument("--pmax", "--p-max",default=None, nargs="?",
                        dest="p_max",
                        help="The maximum probability limit value. "
                             "This determines the cbar limits and contour levels.",
                        type=float)
    optional.add_argument("-pu", "--p-units", "--punits", default="kT", nargs="?",
                        dest="p_units", choices=("kT", "kcal", "raw", "raw_norm"),
                        help="Can be 'kT' (default), 'kcal', 'raw', or 'raw_norm'"
                             "kT = -lnP, kcal/mol = -RT(lnP), where RT=0.5922 at T(298K). "
                             "'raw' is the raw probabilities and "
                             "'raw_norm' is the raw probabilities P(max) normalized.",
                        type=str)
    optional.add_argument("-T", "--temp", default=298, nargs="?",
                        dest="T", help="Used with kcal/mol 'p-units'.",
                        type=int)
    optional.add_argument("-ci", "--contour-interval", default=1, nargs="?",
                        dest="contour_interval",
                        help="If using plot-mode contour, "
                             "This sets the interval of contour level.",
                        type=float)
    optional.add_argument("-cl", "--contour-levels", default=None, nargs="*",
                        dest="contour_levels",
                        help="If using plot-mode contour, "
                             "This overrides and sets the contour levels manually.",
                        type=float)
    optional.add_argument("-sl", "--smoothing-level", default=None, 
                        dest="smoothing_level",
                        help="Smooth data (plotted as histogram or contour"
                             " levels) using a gaussian filter with sigma="
                             "SMOOTHING_LEVEL.",
                        type=float)
    # TODO: is there a better way to do this? 
    #optional.add_argument("--weighted", default=True, action="store_true",
    #                      help="Use weights from WE.")
    optional.add_argument("-nw", "--not-weighted",
                          help="Include this to not use WE weights.",
                          dest="not_weighted", action="store_true")
    # optional.add_argument("--weighted", default=True, 
    #                       action=argparse.BooleanOptionalAction)
    # * args is flexible number of values, which will be gathered into a list
    optional.add_argument("-sb", "--skip-basis", default=None, nargs="*",
                          dest="skip_basis",
                          help="List of binary values for skipping basis states, "
                               "e.g. 0 1 1 to skip all bstates except for first.",
                          type=int)
    optional.add_argument("-jp", "--jointplot", "--joint-plot", default=False,
                          dest="jointplot",
                          help="Optionally include marginal plots to create "
                               "a joint plot from 2D pdist.",
                          action="store_true")
    optional.add_argument("-3d", "--proj3d", default=False,
                          dest="proj3d",
                          help="Make a 3d projection plot",
                          action="store_true")
    # create optional flag to output everything to console screen
    # optional.add_argument("-ots", "--output_to_screen", default=True,
    #                     dest = "output_to_screen",
    #                     help = "Outputs plot to screen. True (default) or False", 
    #                     action= "store_true") 
    optional.add_argument("-nots", "--no-output-to-screen",
                        dest = "no_output_to_screen",
                        help = "Include this argument to not output the plot to "
                        "your display.", 
                        action= "store_true") 
    optional.add_argument("-npb", "--no-progress-bar",
                        dest = "no_pbar",
                        help = "Include this argument to not output the tqdm progress bar.",
                        action= "store_true")

    # plot tracing arg group
    trace = parser.add_argument_group("Optional Plot Tracing", 
                                       description="Plot a trace on top of the pdist.")
    trace_group = trace.add_mutually_exclusive_group()
    # type to float for val inside tuple, 
    # and nargs to 2 since it is interpreted as a 2 item tuple or list
    trace_group.add_argument("--trace-seg", default=None, nargs=2,
                             dest="trace_seg",
                             help="Trace and plot a single continuous trajectory based "
                                 "off of 2 space-seperated ints : iteration segment",
                             type=int)
    trace_group.add_argument("--trace-val", default=None, nargs=2,
                             dest="trace_val",
                             help="Trace and plot a single continuous trajectory based "
                                  "off of 2 space-seprated floats : Xvalue Yvalue",
                             type=float)

    ##########################################################
    ############### FORMATTING ARGUMENTS #####################
    ##########################################################

    formatting = parser.add_argument_group("Plot Formatting Arguments") 

    formatting.add_argument("--style", default="default", nargs="?",
                        dest="style",
                        help="mpl style, can leave blank to use default, "
                             "input `None` for basic mpl settings, can use a custom "
                             "path to a mpl.style text file, or could use a mpl included "
                             "named style, e.g. `ggplot`. "
                             "Edit the wedap/styles/default.mplstyle file to "
                             "change default wedap plotting style options.",
                        type=str)
    # TODO: prob cant use custom outside of list
    formatting.add_argument("--cmap", default="viridis", nargs="?",
                        dest="cmap", help="mpl colormap name.", type=str)
    formatting.add_argument("--color",
                        dest="color", help="Color for 1D plots, contour lines, and trace plots.",
                        widget="ColourChooser")
    formatting.add_argument("--xlabel", dest="xlabel", type=str)
    formatting.add_argument("--xlim", help="LB UB", dest="xlim", nargs=2, type=float)
    formatting.add_argument("--ylabel", dest="ylabel", type=str)
    formatting.add_argument("--ylim", help="LB UB", dest="ylim", nargs=2, type=float)
    formatting.add_argument("--title", dest="title", type=str)
    formatting.add_argument("--suptitle", dest="suptitle", type=str)
    formatting.add_argument("--cbar-label", dest="cbar_label", type=str)
    formatting.add_argument("--grid", dest="grid", default=False, action="store_true")

    # return the argument parser
    return parser 


# TODO: adjust all to fix str/int/type auto
def handle_command_line(argument_parser): 
    """
    Take command line arguments, check for issues, return the arguments. 

    Parameters
    ----------
    argument_parser : argparse.ArgumentParser 
        The argument parser that is returned in `create_cmd_arguments()`.
    
    Returns
    -------
    argparse.NameSpace
        Contains all arguments passed into wedap.
    
    Raises
    ------  
    Prints specific issues to terminal.
    """
    # retrieve args
    args = argument_parser.parse_args() 

     # TODO: check make sure that -jp and -3d not being using at same time

    # h5 file and file exists
    # if not os.path.exists(args.file) or not ".h5" in args.file:  
    #     # print error message and exits
    #     sys.exit("Must input file that exists and is in .h5 file format.")

    # if not args.percentage.isdigit():  # not correct input   
    #     # print out any possible issues
    #     print("You must input a percentage digit. EXAMPLES: \
    #     \n '-p 40' \n You CANNOT add percent sign (eg. 50%) \n You \
    #     CANNOT add decimals (eg. 13.23)") 
    #     sys.exit(0) # exit program

    # # ignore if NoneType since user doesn't want --maxEnsembleSize parameter
    # if args.max_ensemble_size is None: 
    #     pass

    # elif not args.max_ensemble_size.isdigit(): # incorrect input 
    #     # needs to be whole number
    #     print("You must input a whole number with no special characters (eg. 4).")  
    #     sys.exit(0) # exit program 

    # elif args.max_ensemble_size is '0': # user gives 0 to --maxEnsembleSize flag 
    #     print("You cannot input '0' to --maxEnsembleSize flag.")
    #     sys.exit(0) # exit program 

    return args


"""
Unit and regression tests for the H5_Plot class.
"""

# Import package, test suite, and other packages as needed
import wedap

import numpy as np
import pytest

import matplotlib
matplotlib.use('agg')

# look at file coverage for testing
# pytest -v --cov=wedap
# produces .coverage binary file to be used by other tools to visualize 
# do not need 100% coverage, 80-90% is very high

# can have report in 
# $ pytest -v --cov=wedap --cov-report=html
# index.html to better visualize the test coverage

# decorator to skip in pytest
#@pytest.mark.skip

def plot_data_gen(h5, data_type, plot_mode, Xname, Yname=None, Zname=None, out=None):
    """
    Make plot and return or convert to npy binary data file.
    """
    plot = wedap.H5_Plot(h5=h5, data_type=data_type, plot_mode=plot_mode, 
                         Xname=Xname, Yname=Yname, Zname=Zname)
    plot.plot()
    fig = plot.fig

    # draw the figure 
    # tight since canvas is large
    fig.tight_layout(pad=0)
    fig.canvas.draw()

    # Now we can save it to a numpy array.
    data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    
    # save as np binary file
    if out:
        np.save(out, data)
    
    return data

class Test_H5_Plot():
    """
    Test each method of the H5_Plot class.
    """
    h5 = "wedap/data/p53.h5"
    
    @pytest.mark.parametrize("data_type", ["evolution", "average", "instant"])
    @pytest.mark.parametrize("Xname", ["pcoord", "dihedral_2"])
    def test_1_dataset_plots(self, data_type, Xname):
        # make plot data array
        plotdata = plot_data_gen(self.h5, data_type=data_type, plot_mode="line", Xname=Xname)

        # compare to previously generated plot data
        data = np.load(f"wedap/data/plot_{data_type}_line_{Xname}.npy")
        #np.testing.assert_allclose(plotdata, data)
        # check to see if the amount of mismatches is less than 500 (<1% of 1 million items)
        assert data.size - np.count_nonzero(plotdata==data) < 500
        
    @pytest.mark.parametrize("data_type", ["average", "instant"])
    @pytest.mark.parametrize("plot_mode", ["hist", "contour"])
    @pytest.mark.parametrize("Xname, Yname", [["pcoord", "dihedral_2"], ["dihedral_2", "pcoord"]])
    def test_2_dataset_plots(self, data_type, plot_mode, Xname, Yname):
        # make plot data array
        plotdata = plot_data_gen(self.h5, data_type=data_type, plot_mode=plot_mode, 
                                 Xname=Xname, Yname=Yname)

        # compare to previously generated plot data
        data = np.load(f"wedap/data/plot_{data_type}_{plot_mode}_{Xname}_{Yname}.npy")
        #np.testing.assert_allclose(plotdata, data)
        # check to see if the amount of mismatches is less than 500 (<1% of 1 million items)
        assert data.size - np.count_nonzero(plotdata==data) < 500
    
    @pytest.mark.parametrize("data_type", ["average", "instant"])
    @pytest.mark.parametrize("Xname, Yname, Zname", [["pcoord", "dihedral_2", "dihedral_3"], 
                                                     ["dihedral_2", "pcoord", "dihedral_3"], 
                                                     ["dihedral_2", "dihedral_3", "pcoord"]])
    def test_3_dataset_plots(self, data_type, Xname, Yname, Zname):
        # make plot data array
        plotdata = plot_data_gen(self.h5, data_type=data_type, plot_mode="scatter3d", 
                                 Xname=Xname, Yname=Yname, Zname=Zname)

        # compare to previously generated plot data
        data = np.load(f"wedap/data/plot_{data_type}_scatter3d_{Xname}_{Yname}_{Zname}.npy")
        #np.testing.assert_allclose(plotdata, data)
        # check to see if the amount of mismatches is less than 500 (<1% of ~1 million items)
        assert data.size - np.count_nonzero(plotdata==data) < 500



from setuptools import setup, find_packages

with open('./README.md') as f:
    readme = f.read()

with open('./LICENSE') as f:
    lic = f.read()

packages = find_packages()

setup(
    name='wedap',
    version='0.0.7',
    description='weighted ensemble data analysis and plotting',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Darian T. Yang',
    author_email='dty7@pitt.edu',
    install_requires=['numpy', 'matplotlib', 'h5py', 'scipy', 'gif', 'tqdm'],
    url='https://github.com/darianyang/wedap',
    project_urls={'Documentation' : 'https://darianyang.github.io/wedap'},
    license=lic,
    #packages = find_packages(where = 'src'),
    #Packages=find_packages(exclude="docs"),
    packages=find_packages(exclude="docs"),
    package_data={"wedap": ["styles/*"]},
    #py_modules=["wedap"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Chemistry"
    ],
    entry_points={"console_scripts" : ["wedap=wedap.__main__:main",
                                       "mdap=mdap.__main__:main"]}
)



"""
Main plotting class of wedap.
Plot all of the datasets generated with H5_Pdist.

* line -- plot 1D lines.
* hist -- plot histogram (default).
* hist_l -- plot histogram and contour lines.
* contour -- plot contour levels and lines.
* contour_f -- plot contour levels
* contour_l -- plot contour lines only.
* scatter3d -- plot 3 datasets in a scatter plot.

maybe a ridgeline plot?
- This would be maybe for 1D avg of every 100 iterations
- https://matplotlib.org/matplotblog/posts/create-ridgeplots-in-matplotlib/
Option to overlay different datasets, could be done easily with python but maybe a cli option?                

TODO: plot clustering centroids option?
      can then grab the search_aux at the centroid

TODO: bin visualizer? and maybe show the trajectories as just dots?
"""

import sys
import numpy as np
import matplotlib.pyplot as plt
import scipy.ndimage
from warnings import warn
from numpy import inf

from .h5_pdist import *

# TODO: maybe put the pdist object into the plot class and have this object be flexible
# so it could just be a pdist.h5 file from westpa or make your own
# or read in a pdist.h5 and create pdist object using that dataset

class H5_Plot(H5_Pdist):
    """
    These methods provide various plotting options for pdist data.
    """
    def __init__(self, X=None, Y=None, Z=None, plot_mode="hist", cmap=None, smoothing_level=None,
        color=None, ax=None, p_min=None, p_max=None, contour_interval=1, contour_levels=None,
        cbar_label=None, cax=None, jointplot=False, data_label=None, proj3d=False, 
        *args, **kwargs):
        """
        Plotting of pdists generated from H5 datasets.

        Parameters
        ----------
        X, Y : arrays
            x and y axis values, and if using aux_y or evolution (with only aux_x), also must input Z.
        Z : 2darray
            Z is a 2-D matrix of the normalized histogram values.
        plot_mode : str
            TODO: update and expand. Can be 'hist' (default), 'contour', 'line', 'scatter3d'.
        cmap : str
            Can be string or cmap to be input into mpl. Default = viridis.
        smoothing_level : float
            Optionally add gaussian noise to smooth Z data. A good value is around 0.4 to 1.0.
        color : str
            Color for 1D plots.
        ax : mpl axes object
        plot_options : kwargs dictionary
            Include mpl based plot options (e.g. xlabel, ylabel, ylim, xlim, title).
        p_min : int
            The minimum probability limit value.
        p_max : int
            The maximum probability limit value.
        contour_interval : int
            Interval to put contour levels if using 'contour' plot_mode.
        cbar_label : str
            Label for the colorbar.
        cax : MPL axes object
            Optionally define axes object to place colorbar.
        jointplot : bool
            Whether or not to include marginal plots. Note to use this argument, 
            probabilities for Z or from H5_Pdist must be in `raw` p_units.
        data_label : str
            Optionally label the data, e.g. for multiple 1D plots.
        ** args
        ** kwargs
        """
        # include the init args for H5_Pdist
        # TODO: how to make some of the args optional if I want to use classes seperately?
        #super().__init__(*args, **kwargs)

        self.ax = ax
        self.smoothing_level = smoothing_level
        self.jointplot = jointplot

        # TODO: option if you want to generate pdist
        # also need option of just using the input X Y Z args
        # or getting them from w_pdist h5 file, or from H5_Pdist output file
        # user inputs XYZ
        if X is None and Y is None and Z is None:
            super().__init__(*args, **kwargs)
            # save the user requested p_units and changes p_units to raw
            if self.jointplot:
                self.requested_p_units = self.p_units
                kwargs["p_units"] = "raw"
            # will be re-normed later on
            X, Y, Z = H5_Pdist(*args, **kwargs).pdist()
        # need to set this when using mdap, shouldn't affect anything else
        # since joint plot dists must be changed from raw to requested p_units
        if self.jointplot and "p_units" in kwargs:
            self.requested_p_units = kwargs["p_units"]
        else:
            self.requested_p_units = "kT"

        self.X = X
        self.Y = Y
        self.Z = Z

        self.p_min = p_min
        self.p_max = p_max
        self.contour_interval = contour_interval
        self.contour_levels = contour_levels

        self.plot_mode = plot_mode
        self.cmap = cmap
        self.color = color # 1D color
        #self.plot_options = plot_options

        # set cbar_label to default to blank if None
        if cbar_label is None:
            self.cbar_label = ""

        # if no p_units are there then no label is fine
        # otherwise check if p_units are there
        if hasattr(self, "p_units"):
            if self.p_units == "kT":
                self.cbar_label = "$-\ln\,P(x)$"
                #self.cbar_label = "-ln (P(x))"
            elif self.p_units == "kcal":
                self.cbar_label = "$-RT\ \ln\, P\ (kcal\ mol^{-1})$"
            elif self.p_units == "raw":
                self.cbar_label = "Counts"
            elif self.p_units == "raw_norm":
                self.cbar_label = "Normalized Counts"
        # if using 3 datasets, put blank name as default cbar
        if self.plot_mode == "scatter3d":
            self.cbar_label = ""
        # overwrite and apply cbar_label attr if available/specified
        if cbar_label:
            self.cbar_label = cbar_label

        self.cax = cax
        self.data_label = data_label
        self.proj3d = proj3d
        self.kwargs = kwargs

    # TODO: load from w_pdist, also can add method to load from wedap pdist output
    # def _load_from_pdist_file(self):
    #     '''
    #     Load data from a w_pdist output file. This includes bin boundaries. 
    #     '''
    #     # Open the HDF5 file.
    #     self.pdist_HDF5 = h5py.File(self.args.pdist_file)

    #     # Load the histograms and sum along all axes except those specified by
    #     # the user.  Also, only include the iterations specified by the user.
    #     histogram      = numpy.array(self.pdist_HDF5['histograms'])

    #     # Figure out what iterations to use
    #     n_iter_array   = numpy.array(self.pdist_HDF5['n_iter'])
    #     if self.args.first_iter is not None:
    #         first_iter = self.args.first_iter
    #     else:
    #         first_iter = n_iter_array[0] 
    #     if self.args.last_iter is not None:
    #         last_iter = self.args.last_iter
    #     else:
    #         last_iter = n_iter_array[-1]
    #     first_iter_idx = numpy.where(n_iter_array == first_iter)[0][0]
    #     last_iter_idx  = numpy.where(n_iter_array == last_iter)[0][0]
    #     histogram      = histogram[first_iter_idx:last_iter_idx+1]

    #     # Sum along axes
    #     self.axis_list = self._get_bins_from_expr(self.args.pdist_axes)
    #     self.H         = self._sum_except_along(histogram, self.axis_list) 

    #     # Make sure that the axis ordering is correct.
    #     if self.axis_list[0] > self.axis_list[1]:
    #         self.H = self.H.transpose()

    def add_cbar(self, cax=None):
        """
        Add cbar.

        Parameters
        ----------
        cax : mpl cbar axis
            Optionally specify the cbar axis.
        """
        # fig vs plt should be the same, tests run fine (needed to go plt for mosaic)
        #cbar = self.fig.colorbar(self.plot, cax=cax)
        cbar = plt.colorbar(self.plot_obj, cax=cax)
        # if contour lines are present
        if hasattr(self, "lines"):
            cbar.add_lines(self.lines)

        # TODO: move labelpad here to style?
        cbar.set_label(self.cbar_label, labelpad=14)

        # allow for cbar object manipulation (e.g. removal in movie)
        self.cbar = cbar
    
    def plot_hist(self):
        """
        2d hist plot.
        """
        # 2D heatmaps
        # TODO: westpa makes these the max to keep the pdist shape
        # if self.p_max:
        #     self.Z[self.Z > self.p_max] = inf
        self.plot_obj = self.ax.pcolormesh(self.X, self.Y, self.Z, cmap=self.cmap,
                                           shading="auto", vmin=self.p_min, vmax=self.p_max)

    def _get_contour_levels(self):
        """
        Get contour level attribute.
        """
        # TODO: could clean up this logic better
        if self.p_min is None:
            self.p_min = np.min(self.Z)
        # if levels aren't specified
        if self.contour_levels is None:
            # 2D contour plots
            if self.p_max is None:
                warn("With 'contour' plot_type, p_max should be set. Otherwise max Z is used.")
                self.contour_levels = np.arange(self.p_min, np.max(self.Z[self.Z != np.inf ]), self.contour_interval)
            elif self.p_max <= 1:
                warn("You may want to change the `contour_interval` argument to be < 1")
                self.contour_levels = np.arange(self.p_min, self.p_max + self.contour_interval, self.contour_interval)
            else:
                self.contour_levels = np.arange(self.p_min, self.p_max + self.contour_interval, self.contour_interval)

    def plot_contour_l(self):
        """
        2d contour plot, lines.
        """
        Warning("contour_l lines are set to mpl defaults, set can be changed with `--color` or `--cmap`")
        # can control linewidths using rc params (lines.linewidths (default 1.5))
        if self.color:
            self.lines = self.ax.contour(self.X, self.Y, self.Z, levels=self.contour_levels, colors=self.color)
            #self.lines = self.ax.contour(self.X, self.Y, self.Z, levels=[5], colors=self.color)
        else:
            self.lines = self.ax.contour(self.X, self.Y, self.Z, levels=self.contour_levels, cmap=self.cmap)

    def plot_contour_f(self):
        """
        2d contour plot, fill.
        """
        self.plot_obj = self.ax.contourf(self.X, self.Y, self.Z, levels=self.contour_levels, cmap=self.cmap)

    def plot_bar(self):
        """
        Simple bar plot.
        """
        # 1D data
        self.ax.bar(self.X, self.Y, color=self.color, label=self.data_label)
        self.ax.set_ylabel("P(x)")

    def plot_line(self):
        """
        1d line plot.
        """
        # 1D data
        if self.p_max:
            self.Y[self.Y > self.p_max] = inf
        self.ax.plot(self.X, self.Y, color=self.color, label=self.data_label)
        self.ax.set_ylabel(self.cbar_label)
    
    def plot_scatter3d(self, interval=10, s=1):
        """
        3d scatter plot.

        Parameters
        ----------
        interval : int
            Interval to consider the XYZ datasets, increase to use less data.
        s : float
            mpl scatter marker size.
        """
        if self.proj3d:
            self.plot_obj = self.ax.scatter(self.X[::interval], 
                                            self.Y[::interval], 
                                            self.Z[::interval],
                                            c=self.Z[::interval],
                                            cmap=self.cmap, s=s,
                                            vmin=self.p_min, vmax=self.p_max)
        else:
            self.plot_obj = self.ax.scatter(self.X[::interval], 
                                            self.Y[::interval], 
                                            c=self.Z[::interval], 
                                            cmap=self.cmap, s=s,
                                            vmin=self.p_min, vmax=self.p_max)

    def plot_hexbin3d(self):
        """
        Hexbin plot?
        """
        # TODO: test this and add grid?
        self.plot_obj = self.ax.hexbin(self.X, self.Y, C=self.Z,
                                       cmap=self.cmap, vmin=self.p_min, vmax=self.p_max)

    def plot_margins(self):
        """
        Joint plot of heatmap (pcolormesh).
        Must input raw probabilities from H5_Pdist(p_units = 'raw').
        """
        # clean up infs and NaNs in Z
        Z = np.ma.masked_invalid(self.Z)
        # calc margin datasets
        x_proj = Z.sum(axis=0)
        y_proj = Z.sum(axis=1)
        # plot margins
        self.fig["x"].plot(self.X, self._normalize(x_proj), color=self.color)
        self.fig["y"].plot(self._normalize(y_proj), self.Y, color=self.color)

        # TODO: add functionality for scatter3d, use XY data to create gaussian_kde margins

    def _unpack_plot_options(self):
        """
        Unpack the plot_options kwarg dictionary.
        """
        # unpack plot options dictionary making sure not None
        # TODO: put all in ax.set()?
        #for key, item in self.plot_options.items():
        for key, item in self.kwargs.items():
            if key == "xlabel" and item:
                self.ax.set_xlabel(item)
            if key == "ylabel" and item:
                self.ax.set_ylabel(item)
            if key == "xlim" and item:
                self.ax.set_xlim(item)
                if self.jointplot:
                    self.fig["x"].set_xlim(item)
            if key == "ylim"and item:
                self.ax.set_ylim(item)
                if self.jointplot:
                    self.fig["y"].set_ylim(item)
            if key == "title" and item:
                self.ax.set_title(item)
            if key == "suptitle" and item:
                plt.suptitle(item)
            if key == "grid" and item:
                self.ax.grid(item, alpha=0.5)
                if self.jointplot:
                    # grid the margins
                    for ax in ["x", "y"]:
                        self.fig[ax].grid(item, alpha=0.5)
            if key == "minima" and item: # TODO: this is essentially bstate, also put maxima?
                # reorient transposed hist matrix
                Z = np.rot90(np.flip(self.Z, axis=0), k=3)
                # get minima coordinates index (inverse maxima since min = 0)
                maxima = np.where(1 / Z ==  np.amax(1 / Z, axis=(0, 1)))
                # plot point at x and y bin midpoints that correspond to mimima
                self.ax.plot(self.X[maxima[0]], self.Y[maxima[1]], 'ko')
                print(f"Minima: ({self.X[maxima[0]][0]}, {self.Y[maxima[1]][0]})")

    # TODO: cbar issues with 1d plots
    def plot(self, cbar=True):
        """
        Main public method.
        Master plotting run function
        Parse plot type and add cbars/tightlayout/plot_options/smoothing

        Parameters
        ----------
        cbar : bool
            Whether or not to include a colorbar.
        """
        # special settings for joint plots
        if self.jointplot:
            # can't use custom ax objects with jointplot, must create new fig and ax using mosaic
            if self.ax:
                message = "Can't use custom mpl axes objects with jointplot option, " + \
                          "creating new fig and ax using mpl subplot_mosaic."
                warn(message)
            # right now can't handle scatter with joint plot, it wouldn't be kT but standard hist
            if self.plot_mode == "scatter3d":
                warn("EXITING: Currently can't use `--plot-mode scatter3d` with `--jointplot`.")
                sys.exit(0)
            # since jointplot starts with raw probabilities
            # need to figure out what p_units are needed
            # only if p_units is definied (e.g. H5_Pdist args are in place)
            try:
                self.p_units
                self.T
            # if H5_Pdist args not in place, use default
            # TODO: this warning always pops up even is args.p_units is set
            except AttributeError as e:
                warn(f"{e}: Defaulting to 'kT' probability units.")
                # self.p_units does not exist, default to kT
                self.p_units = "kT"
            # if H5_Pdists args were in place and changed to "raw" for jointplots
            if self.p_units == "raw":
                # return to requested p_units
                self.p_units = self.requested_p_units

            # set figure attribute to be a mosaic plot
            self.fig = plt.figure(layout="tight").subplot_mosaic(
                                                    """
                                                    x..
                                                    Hyc
                                                    """,
                                                    # set the height ratios between the rows
                                                    height_ratios=[1, 3.5],
                                                    # set the width ratios between the columns
                                                    width_ratios=[3.5, 1, 0.25],
                                                    )
            self.cax = self.fig["c"]
            self.ax = self.fig["H"]
            # plot margins first from raw probabilities
            self.plot_margins()
            # calc normalized hist using updated p_units (maybe could also do Z[Z != np.isinf])
            # but masked invalid takes care of Nan and inf converts to mask (--)
            #self.Z = self._normalize(self.Z[self.Z != np.inf])
            self.Z = self._normalize(np.ma.masked_invalid(self.Z))

            # add formatting jointplot options
            # remove redundant tick labels
            # self.fig["x"].set_xticks([])
            # self.fig["y"].set_yticks([])
            self.fig["x"].set_xticklabels([])
            self.fig["y"].set_yticklabels([])
            # put pmax as lims on margin plots
            self.fig["x"].set_ylim(self.p_min, self.p_max)
            self.fig["y"].set_xlim(self.p_min, self.p_max)

        # 3dprojection test, not going to be compatible with jp
        elif self.proj3d:
            self.fig = plt.figure()
            self.ax = self.fig.add_subplot(projection='3d')
            # don't need cbar since 3d projected
            cbar = False
            # but add cbar label to z axis
            self.ax.set_zlabel(self.cbar_label)

        else:
            if self.ax is None:
                self.fig, self.ax = plt.subplots()
            else:
                self.fig = plt.gcf()

        # smooth the data if specified
        if self.smoothing_level:
            # make into a smooth style e.g. contour with 0 as base instead of negative or inf
            #self.Z = np.ma.masked_invalid(self.Z)
            self.Z[self.Z < 0] = 0
            self.Z[self.Z == np.inf] = 0
            self.Z = scipy.ndimage.gaussian_filter(self.Z, sigma=self.smoothing_level)
            #self.Z[self.Z == 0] = np.inf
            # get rid of any negatives --> 0
            #self.Z[self.Z < 0] = np.inf
            #self.Z[self.Z == np.inf] = 0
            #self.Z[self.Z < 0] = 0

        # get contour levels if needed
        if self.plot_mode in ["contour", "contour_l", "contour_f", "hist_l"]:
            self._get_contour_levels()

        if self.plot_mode == "contour":
            self.plot_contour_l()
            self.plot_contour_f()
        elif self.plot_mode == "contour_l":
            self.plot_contour_l()
        elif self.plot_mode == "contour_f":
            self.plot_contour_f()

        # TODO: auto label WE iterations on evolution? (done via __main__ right now)
        elif self.plot_mode == "hist" or self.plot_mode == "hist_l":
            if self.plot_mode == "hist_l":
                self.plot_contour_l()
            # I run into this error when I run something like instant with 
            # the h5 but didn't adjust the plot mode to something like line
            try:
                self.plot_hist()
            except (TypeError,ValueError) as e:
                # TODO: put the text into logger?
                print(f"{e}: Did you mean to use the default 'hist' plot mode?")
                print("Perhaps you need to define another dimension via '--Yname'?")
                print("Or if you want a 1D pdist, set --plot-mode or -pm to `line`.")
                sys.exit()

        elif self.plot_mode == "bar":
            self.plot_bar()
            Warning("'bar' plot_mode is still under development")

        elif self.plot_mode == "line":
            self.plot_line()
            #self.ax.set_ylabel(self.cbar_label)

        elif self.plot_mode == "scatter3d":
            self.plot_scatter3d()

        elif self.plot_mode == "hexbin3d":
            self.plot_hexbin3d()

        # error if unknown plot_mode
        else:
            raise ValueError(f"plot_mode = '{self.plot_mode}' is not valid.")

        # TODO: can this work with non H5_Pdist input?
        # if self.Xname == "pcoord":
        #     self.ax.set_xlabel(f"Progress Coordinate {self.Xindex}")
        # if self.Yname == "pcoord":
        #     self.ax.set_ylabel(f"Progress Coordinate {self.Yindex}")

        # don't add cbar if not specified or if using a 1D plot
        if cbar and self.plot_mode not in ["line", "bar", "contour_l"]:
            self.add_cbar(cax=self.cax)

        # take kwargs and unpack to look for plot option items
        if self.kwargs is not None:
            self._unpack_plot_options()

        # fig vs plt shouldn't matter here (needed to go plt for mosaic)
        #self.fig.tight_layout()
        plt.tight_layout()



"""
Convert auxillary data recorded during WESTPA simulation and stored in west.h5 file
to various probability density datasets.

This script effectively replaces the need to use the native WESTPA plotting pipeline:
west.h5 --w_pdist(with --construct-dataset module.py)--> 
pdist.h5 --plothist(with --postprocess-functions hist_settings.py)--> plot.pdf

TODO: 
    - maybe add option to output pdist as file, this would speed up subsequent plotting
        of the same data. H5_Plot could then use this data.
    - add option for a list of equivalent h5 files, alternative to w_multi_west.
"""

# TEMP for trace plot (TODO)
import matplotlib.pyplot as plt
######

import h5py
import numpy as np
from tqdm.auto import tqdm

from scipy.spatial import KDTree

from warnings import warn

# for copying h5 file
import shutil

# Suppress divide-by-zero in log
np.seterr(divide='ignore', invalid='ignore')

class H5_Pdist():
    """
    These class methods generate probability distributions from a WESTPA H5 file.
    """
    # TODO: is setting aux_y to None the best approach to 1D plot settings?
    # TODO: add step-iter
    def __init__(self, h5="west.h5", data_type=None, Xname="pcoord", Xindex=0, Yname=None, 
                 Yindex=0, Zname=None, Zindex=0, H5save_out=None, Xsave_name=None, Ysave_name=None,
                 Zsave_name=None, data_proc=None, first_iter=1, last_iter=None, bins=(100,100), 
                 p_units='kT', T=298, weighted=True, skip_basis=None, skip_basis_out=None,
                 histrange_x=None, histrange_y=None, no_pbar=False, *args, **kwargs):
        """
        Parameters
        ----------
        h5 : str
            path to west.h5 file.
        data_type : str
            'evolution' (1 dataset); 'average' or 'instant' (1 or 2 datasets)
        Xname : str
            target data for x axis, default pcoord.
        Xindex : int
            If X.ndim > 2, use this to index.
        Yname : str
            target data for y axis, default None.
        Yindex : int
            If Y.ndim > 2, use this to index.
        Zname : str
            target data for z axis, default None. 
            Use this if you want to use a dataset instead of pdist for Z axis.
            This will be best plotted as a scatter plot with Z as the marker color.
            Instead of returning the pdist, only the XYZ datasets will be returned.
            This is becasue the weights/pdist isn't considered.
        Zindex : int
            If Z.ndim > 2, use this to index.
        H5save_out : str
            Paths to save a new H5 file with this dataset name.
            Right now it saves the requested X Y or Z data into a new aux_name.
            Note if you use this feature the input data must be the same shape and formatting as the other
            H5 file datasets. (TODO: organization?)
        Xsave_name, Ysave_name, Zsave_name : str
            Respective names to call the new dataset saved into the new H5 file.
        data_proc : function or tuple of functions
            Of the form f(data) where data has rows=segments, columns=frames until tau, depth=data dims.
            The input function must return a processed array of the same shape and formatting.
        first_iter : int
            Default start plot at iteration 1 data.
        last_iter : int
            Last iteration data to include, default is the last recorded iteration in the west.h5 file. 
            Note that `instant` type pdists only depend on last_iter.
        bins : tuple of ints (TODO: maybe the tuple isn't user friendly for 1 dim?)
            Histogram bins in pdist data to be generated for x and y datasets, default both 100.
        p_units : str
            Can be 'kT' (default), 'kcal', 'raw', or 'raw_norm'.
            kT = -lnP, kcal/mol = -RT(lnP), where RT = 0.5922 at `T` Kelvin.
            'raw' is the raw probabilities and 'raw_norm' is the raw probabilities P(max) normalized.
        T : int
            Temperature if using kcal/mol.
        weighted : bool
            Default True, use WE segment weights in pdist calculation.
        skip_basis : list
            List of binaries for each basis state to determine if it is skipped.
            e.g. [0, 0, 1] would only consider the trajectory data from basis 
            states 1 and 2 but would skip basis state 3, applying zero weights.
        skip_basis_out : str
            Name of the outfile h5 file for optionally outputting new skipped basis h5 dataset.
        histrange_x, histrange_y : list or tuple of 2 floats or ints
            Optionally put custom bin ranges.
        no_pbar : bool
            Optionally do not include the progress bar for pdist generation.
        TODO: maybe also binsfromexpression?
        """
        # TODO: maybe change self.f to self.h5?
        self.h5 = str(h5)
        self.f = h5py.File(h5, mode="r")
        if data_type is None:
            raise ValueError("Must input valid data_type: `evolution`, `average`, or `instant`")
        else:
            self.data_type = data_type
        self.p_units = str(p_units)
        self.T = int(T)
        self.weighted = weighted

        # TODO: Default pcoord for either dim
        # TODO: clean up and condense this name processing section
        # add auxdata prefix if not using pcoord and not using array input
        # doing it in two conditional blocks since numpy as warning with comparing array to string
        # this way it only does the string comparison if Xname is a string
        if isinstance(Xname, str):
            if Xname != "pcoord":
                Xname = "auxdata/" + Xname
        self.Xname = Xname
        # TODO: set this up as an arg to be able to process 3D+ arrays form aux
        # need to define the index if pcoord is 3D+ array, index is ndim - 1
        self.Xindex = Xindex

        # for 1D plots, but could be better (TODO)
        if Yname is not None and isinstance(Yname, str):
            # for common case of evolution with extra Yname input
            if Yname and data_type == "evolution":
                message = "\nDefaulting to evolution plot for --data-type, since you put a --Yname arg, " + \
                          "\nDid you mean to use --data-type of `average` or `instant`?"
                warn(message)
            # add auxdata prefix if not using pcoord and not using array input
            if Yname != "pcoord":
                Yname = "auxdata/" + Yname
            # before comparing X and Y, make sure they are both strings
            if isinstance(Xname, str):
                # for the common case where one plots pcoord/aux 0 and pcoord/aux 1
                if Xname == Yname and Xindex == 0 and Yindex == 0:
                    Yindex = 1
                    message = "\nSetting --Yindex to 1 (2nd dimension) since Xname/Yname " + \
                              "and Xindex/Yindex were the same."
                    warn(message)
        self.Yname = Yname
        self.Yindex = Yindex

        # for replacing the Z axis pdist with a dataset
        if Zname is not None and isinstance(Zname, str):
            # add auxdata prefix if not using pcoord and not using array input
            if Zname != "pcoord":
                Zname = "auxdata/" + Zname
        self.Zname = Zname
        self.Zindex = Zindex

        # XYZ save into new h5 file options
        self.H5save_out = H5save_out
        self.Xsave_name = Xsave_name
        self.Ysave_name = Ysave_name
        self.Zsave_name = Zsave_name
        if H5save_out is not None:
            shutil.copyfile(self.h5, H5save_out)
            self.H5save_out = h5py.File(H5save_out, "r+")
        
        # raw data processing function
        # TODO: allow for 2-3 functions as tuple input, right now one function only
        self.data_proc = data_proc

        # default to last
        if last_iter is not None:
            self.last_iter = int(last_iter)
        elif last_iter is None:
            self.last_iter = self.f.attrs["west_current_iteration"] - 1

        if data_type == "instant":
            self.first_iter = self.last_iter
        else:
            self.first_iter = int(first_iter)

        self.bins = bins

        # first make a list for each iteration weight array
        weights = []
        #for iter in range(self.first_iter, self.last_iter + 1):
        # have to make array start from iteration 1 to index well during weighting
        # but only for using skipping basis
        if skip_basis is None:
            weight_start = self.first_iter
        elif skip_basis:
            weight_start = 1
        for iter in range(weight_start, self.last_iter + 1):
            weights.append(self.f[f"iterations/iter_{iter:08d}/seg_index"]["weight"])
        # 1D array of variably shaped arrays
        self.weights = np.array(weights, dtype=object)

        # integer for the amount of frames saved (length) per tau (e.g. 101 for 100 ps tau)
        self.tau = self._get_data_array("pcoord", 0, self.first_iter).shape[1]

        # n_particles for each iteration
        self.n_particles = self.f["summary"]["n_particles"]#[self.first_iter-1:self.last_iter]

        # TODO: I wonder if both of these attributes are needed (total only used by reshape data array)
        #       I should note somewhere that data array must be for the same length/iters as the west.h5 file
        # the sum of n segments in all specified iterations and all iterations overall
        self.current_particles = np.sum(self.f["summary"]["n_particles"][self.first_iter-1:self.last_iter])
        # do not include the final (empty) iteration
        self.total_particles = np.sum(self.f["summary"]["n_particles"][:-1])

        self.skip_basis = skip_basis
        self.skip_basis_out = skip_basis_out
        self.histrange_x = histrange_x
        self.histrange_y = histrange_y
        self.no_pbar = no_pbar

    def _get_data_array(self, name, index, iteration, h5_create=None, h5_create_name=None):
        """
        Extract, index, and return the aux/data array of interest.
        Rows are segments, columns are frames until tau, depth is ndimensional datasets.

        Parameters
        ----------
        name : str
            Dataset name.
        index : int
            Dataset index.
        iteration : int
            WE iteration.
        h5_create : str
            Name of the h5 file to add the dataset to.
        h5_create_name : str
            Name of the dataset that is being placed into the h5 file.

        Returns
        -------
        data : ndarray
            Dataset of interest from the H5 file.
        """
        # if the user puts in an array object instead of a string dataset name
        if isinstance(name, np.ndarray):
            # first reshape 1d input raw data array into 3d array
            # currently, this is done during pdist method
            #data = self.reshape_total_data_array(name)

            # need to parse data for segments only in current iteration
            # segments are each row, but in input they are all concatenated
            n_segs_up_to_iter = np.sum(self.n_particles[self.first_iter-1:iteration-1])
            n_segs_including_iter = np.sum(self.n_particles[self.first_iter-1:iteration])
            data = name[n_segs_up_to_iter:n_segs_including_iter,:,:]
        # name should be a string for the h5 file dataset name
        elif isinstance(name, str):
            # this t/e block is to catch non-existent aux data names
            try:
                data = np.array(self.f[f"iterations/iter_{iteration:08d}/{name}"])
            except KeyError:
                message = f"{name} is not a valid object in the h5 file. \n" + \
                          f"Available datasets are: 'pcoord' "
                # this t/e block is to catch the case where there are no aux datasets at all
                try:
                    auxnames = list(self.f[f"iterations/iter_{self.first_iter:08d}/auxdata"])
                    message += f"and the following aux datasets {auxnames}"
                except KeyError:
                    message += "and no aux datasets were found"
                raise ValueError(message)
                
        else:
            raise ValueError("Xname Yname and Zname arguments must be either a string or an array.")

        # standardize the data dimensions to allow 3d indexing
        data = np.atleast_3d(data)

        # run data processing function on raw data if available
        # don't do this for input array data
        if self.data_proc is not None and not isinstance(name, np.ndarray):
            data = self.data_proc(data)

        # option to fill out new h5 file with dataset included here
        # using westpa style compression and scaleoffset 
        # this way it gets repacked into h5 file to update storage
        # could round to 4 decimal places with data=np.around(data, 4)
        if h5_create:
            h5_create.create_dataset(f"iterations/iter_{iteration:08d}/auxdata/{h5_create_name}", 
                                     data=data, compression=4, scaleoffset=6, chunks=True)

        return data[:,:,index]

    # this does add a little overhead at high iteration ranges
    # ~0.5s from 100i to 400i
    # alternatively, can put histrange_x and histrange_y args to skip this
    def _get_histrange(self, name, index):
        """
        Get the histrange considering the min/max of all iterations considered.

        Parameters
        ----------
        name : str
            Target auxillary data name for range calculation.
        index : int
            Target auxillary data index for range calculation.

        Returns
        -------
        histrange : tuple
            2 item list of min and max bin bounds for hist range of target aux data.
        """
        # set base histrange based on first iteration
        iter_data = self._get_data_array(name, index, self.first_iter)
        histrange = [np.amin(iter_data), np.amax(iter_data)]

        # loop and update to the max and min for all other iterations considered
        for iter in range(self.first_iter + 1, self.last_iter + 1):
            # get min and max for the iteration
            iter_data = self._get_data_array(name, index, iter)
            iter_min = np.amin(iter_data)
            iter_max = np.amax(iter_data)

            # update to get the largest possible range from all iterations
            if iter_min < histrange[0]:
                histrange[0] = iter_min
            if iter_max > histrange[1]:
                histrange[1] = iter_max

        return histrange

    def _normalize(self, hist):
        """
        Normalize or convert the probabilities.

        Parameters
        ----------
        hist : ndarray
            Array containing the histogram count values to be normalized.

        Returns
        -------
        hist : ndarray
            The hist array is normalized according to the p_units argument. 
        """
        # -lnP
        if self.p_units == "kT":
            hist = -np.log(hist / np.max(hist))
        # -RT*lnP
        elif self.p_units == "kcal":
            # Gas constant R = 1.9872 cal/K*mol or 0.0019872 kcal/K*mol
            hist = -0.0019872 * self.T * np.log(hist / np.max(hist))
        # raw probability
        elif self.p_units == "raw":
            hist = hist
        # raw normalized probability (P(x)/P(max))
        elif self.p_units == "raw_norm":
            hist = hist / np.max(hist)
        else:
            raise ValueError("Invalid p_units value, must be 'kT', 'kcal', 'raw', or 'raw_norm'.")
        return hist

    def _get_children_indices(self, parent):
        """
        For a (iter, seg) pair, look for and return all iter + 1 segment indices.

        Parameters
        ----------
        parent : tuple
            (iteration, segment).

        Returns
        -------
        children : tuple
            The indices of all child segments from the input parent segment.
        """
        children = []
        p_iter, p_seg = parent
        # for all parent_id values of the iter + 1 iteration
        for idx, seg in enumerate(
        self.f[f"iterations/iter_{p_iter+1:08d}/seg_index"]["parent_id"]):
            # match with input parent segment
            if seg == p_seg:
                # put all the parent segment children's indices into list
                children.append(idx)
                
        return children

    def _new_weights_from_skip_basis(self):
        """
        Make a new temp h5 file with zero weights for skipped basis state walkers.

        Returns
        -------
        self.weights : array
            Updated weight array with zero values for skipped basis states.
        """
        # TODO: doesn't work with --first-iter
        
        # copy of weights to edit
        new_weights = self.weights

        # setup a warning for h5 files that have incorrectly recorded bstate pcoords
        # this will all be based off of the first pcoord array (Z index 0)
        # correspondingly, the bstate_pcoord will be the first column
        bs_coords = self.f[f"ibstates/0/bstate_pcoord"]
        it1_coords = self.f[f"iterations/iter_00000001/pcoord"][:,0,0]
        # need to second element get the unique indices
        it1_unique_indices = np.unique(it1_coords, return_index=True)[1]
        # then sort to the original bstate ordering
        it1_unique_coords = np.array([it1_coords[index] \
                            for index in sorted(it1_unique_indices)])
        try:
            # make sure that traced unique pcoord elements match the basis state values
            if np.isclose(bs_coords[:,0], it1_unique_coords, rtol=1e-04) is False:
                message = f"The traced pcoord \n{it1_unique_coords} \ndoes not equal " + \
                          f"the basis state coordinates \n{bs_coords}"
                warn(message)
        except ValueError as e:
            message = f"{e}: Not all bstates may have been used in iteration 1."
            warn(message)

        # TODO: print bstate pcoords
        print(f"bstates: {bs_coords[:,0]}")
        print(f"bstates from pcoord: {it1_unique_coords}")
        #import sys ; sys.exit(0)

        # if the basis state binary is a 1 in skip_basis, use weight 0 
        #print("First run skip_basis processing from each initial segment: ")
        for basis, skip in enumerate(self.skip_basis):
            # essentially goes through all initial segments for each skipped basis
            if skip == 1:
                # loop of each initial pcoord value from iteration 1
                for it1_idx, it1_val in enumerate(it1_coords):
                    # so if the pcoord value matches the bstate value to be skipped
                    # needs to both be at the same precision
                    if np.isclose(it1_val, bs_coords[basis,0], rtol=1e-04):
                        # search forward to look for children of basis state 
                        # start at it1_idx, make weight zero 
                        new_weights[0][it1_idx] = 0

                        # list for parent_ids of the current segment skip basis lineage
                        skip_parents_c = [it1_idx]
                        # list for storing the indices to skip for the next iteration
                        skip_parents_n = []

                        # zero the next iteration's children until last_iter
                        for iter in tqdm(range(1, self.last_iter + 1), 
                                         desc="skip_basis", disable=self.no_pbar):
                            for idx in skip_parents_c:
                                # make zero for each child of skip_basis
                                new_weights[iter-1][idx] = 0
                                # then make new skip_parents tuple to loop for next iter
                                skip_parents_n += self._get_children_indices((iter, idx))

                            # make new empty list to store the iteration's skipped
                            skip_parents_c.clear()
                            skip_parents_c += skip_parents_n
                            skip_parents_n.clear()

        # TODO: prob can do better than these print statements
        print("pdist calculation: ")
        # write new weights into skip_basis_out h5 file
        if self.skip_basis_out is not None:
            shutil.copyfile(self.h5, self.skip_basis_out)
            h5_skip_basis = h5py.File(self.skip_basis_out, "r+")
            for idx, weight in enumerate(new_weights):
                h5_skip_basis[f"iterations/iter_{idx+1:08d}/seg_index"]["weight"] = weight
            
        # only return portion of weights requested by user
        return new_weights[self.first_iter-1:self.last_iter]

    # TODO: clean up and optimize
    def search_aux_xy_nn(self, val_x, val_y):
        """
        Originally adapted from code by Jeremy Leung.
        Tree search to find closest datapoint to input data value(s).

        # TODO: add step size for searching, right now gets the last frame

        Parameters
        ----------
        val_x : int or float
            X dataset value to search for.
        val_y : int or float
            Y dataset value to search for.
        """
        # iter is already known when searching evo data
        if self.data_type == "evolution":
            iter_num = int(val_y)
            
            # These are the auxillary coordinates you're looking for
            r1 = self._get_data_array(self.Xname, self.Xindex, iter_num)[:,-1]

            # phase 2: finding seg number

            # TODO: numpy array this
            small_array2 = []
            for j in range(0,len(r1)):
                small_array2.append([r1[j]])
            tree2 = KDTree(small_array2)

            # TODO: these can be multiple points, maybe can parse these and filter later
            d2, i2 = tree2.query([val_x],k=1)
            seg_num = int(i2)

        else:
            # phase 1: finding iteration number
            distances = []
            indices = []

            # change indices to number of iteration
            #for i in range(self.first_iter, self.last_iter + 1): 
            # always use iteration 1 to get full trace path
            for i in range(1, self.last_iter + 1): 

                # These are the auxillary coordinates you're looking for
                r1 = self._get_data_array(self.Xname, self.Xindex, i)[:,-1]
                r2 = self._get_data_array(self.Yname, self.Yindex, i)[:,-1]

                small_array = []
                for j in range(0,len(r1)):
                    small_array.append([r1[j],r2[j]])
                tree = KDTree(small_array)

                # Outputs are distance from neighbour (dd) and indices of output (ii)
                dd, ii = tree.query([val_x, val_y],k=1) 
                distances.append(dd) 
                indices.append(ii)

            minimum = np.argmin(distances)
            iter_num = int(minimum+1)

            # These are the auxillary coordinates you're looking for
            r1 = self._get_data_array(self.Xname, self.Xindex, iter_num)[:,-1]
            r2 = self._get_data_array(self.Yname, self.Yindex, iter_num)[:,-1]

            # phase 2: finding seg number

            # TODO: numpy array this
            small_array2 = []
            for j in range(0,len(r1)):
                small_array2.append([r1[j],r2[j]])
            tree2 = KDTree(small_array2)

            # TODO: these can be multiple points, maybe can parse these and filter later
            d2, i2 = tree2.query([val_x, val_y],k=1)
            seg_num = int(i2)

        #print("go to iter " + str(iter_num) + ", " + "and seg " + str(seg_num))
        print(f"Go to ITERATION: {iter_num} and SEGMENT: {seg_num}")
        return iter_num, seg_num

    ##################### TODO: update or organize this #############################
    def get_parents(self, walker_tuple):
        """
        Get parent of an input (iteration, walker).

        Parameters
        ----------
        walker_tuple : tuple
            (iteration, walker)

        Returns
        -------
        parent : iteration, walker
        """
        it, wlk = walker_tuple
        parent = self.f[f"iterations/iter_{it:08d}"]["seg_index"]["parent_id"][wlk]
        return it-1, parent

    def trace_walker(self, walker_tuple):
        """
        Get trace path of an input (iteration, walker).

        Parameters
        ----------
        walker_tuple : tuple
            (iteration, walker)

        Returns
        -------
        trace : list of tuples
            Tuples are (iteration, walker) traces.
        """
        # Unroll the tuple into iteration/walker 
        it, wlk = walker_tuple
        # Initialize our path
        path = [(it,wlk)]
        # And trace it
        while it > 1: 
            it, wlk = self.get_parents((it, wlk))
            path.append((it,wlk))
        return np.array(sorted(path, key=lambda x: x[0]))

    def get_coords(self, path, data_name, data_index):
        """
        Get a list of data coordinates for plotting traces.

        Parameters
        ----------
        path : list of tuples
            Tuples are (iteration, walker) traces.
        data_name : str
            Name of dataset.
        data_index : int
            Index of dataset.

        Returns
        -------
        coordinates : array
            Array of coordinates from the list of (iteration, walker) tuples.
        """
        # Initialize a list for the pcoords
        coords = []
        # Loop over the path and get the pcoords for each walker
        for it, wlk in path:
            coords.append(self._get_data_array(data_name, data_index, it)[wlk][::10])
        return np.array(coords)

    def plot_trace(self, walker_tuple, color="white", ax=None):
        """
        Plot trace.

        Parameters
        ----------
        walker_tuple : tuple
            (iteration, walker) start point to trace from.
        color : str
        ax : mpl axes object
        """
        # TODO: update/streamline this
        if ax is None:
            fig, ax = plt.subplots(figsize=(7,5))
        else:
            fig = plt.gcf()

        path = self.trace_walker(walker_tuple)
        # adjustments for plothist evolution of only aux_x data
        if self.data_type == "evolution":
            # split iterations up to provide y-values for each x-value (pcoord)
            aux = self.get_coords(path, self.Xname, self.Xindex)
            iters = np.arange(1, len(aux)+1)
            ax.plot(aux[:,0], iters, c="black", lw=2)
            ax.plot(aux[:,0], iters, c=color, lw=1)
            return

        # And pull aux_coords for the path calculated
        aux_x = self.get_coords(path, self.Xname, self.Xindex)
        aux_y = self.get_coords(path, self.Yname, self.Yindex)

        ax.plot(aux_x[:,0], aux_y[:,0], c="black", lw=2)
        ax.plot(aux_x[:,0], aux_y[:,0], c=color, lw=1)

    def w_succ(self):
        """
        Find and return all successfully recycled (iter, seg) pairs.
        TODO eventually can use this to plot pdist of succ only trajs
        note that I would have to norm by the overall pmax (not just succ pmax)
        Could have this be an optional feature.
        """
        succ = []
        for iter in range(self.last_iter):
            # if the new_weights group exists in the h5 file
            if f"iterations/iter_{iter:08d}/new_weights" in self.h5:
                prev_segs = self.f[f"iterations/iter_{iter:08d}/new_weights/index"]["prev_seg_id"]
                # append the previous iter and previous seg id recycled
                for seg in prev_segs:
                    succ.append((iter-1, seg))
        # TODO: order this by iter and seg vals? currently segs not sorted but is iter ordered
        return succ
    def succ_pdist(self):
        """
        TODO: Filter weights to be zero for all non successfull trajectories.
        Make an array of zero weights and fill out weights for succ trajs only.
        """
        pass
    ###############################################################################

    def aux_to_pdist_1d(self, iteration):
        """
        Take the auxiliary dataset for a single iteration and generate a weighted
        1D probability distribution. 

        Parameters
        ----------
        iteration : int
            Desired iteration to extract timeseries info from.

        Returns
        -------
        midpoints_x : ndarray
            Histogram midpoint bin values for target aux coordinate of dimension 0.
        midpoints_y : ndarray
            Optional histogram midpoint bin values for target aux coordinate of dimension 1.
        histogram : ndarray
            Raw histogram count values of each histogram bin. Can be later normalized as -lnP(x).
        """
        # return 1D aux data: 1D array for histogram and midpoint values
        aux = self._get_data_array(self.Xname, self.Xindex, iteration)

        # make an 1D array to fit the hist values based off of bin count
        histogram = np.zeros((self.bins[0]))
        for seg in range(0, aux.shape[0]):
            counts, bins = np.histogram(aux[seg], bins=self.bins[0], range=self.histrange_x)

            # selectively apply weights
            if self.weighted is True:
                # multiply counts vector by weight scalar from weight array
                counts = np.multiply(counts, self.weights[iteration - self.first_iter][seg])

            # add all of the weighted walkers to total array for the 
            # resulting linear combination
            histogram = np.add(histogram, counts)

        # get bin midpoints
        midpoints_x = (bins[:-1] + bins[1:]) / 2
        
        # TODO: also save as instance attributes
        return midpoints_x, histogram

    def aux_to_pdist_2d(self, iteration):
        """
        Take the auxiliary dataset for a single iteration and generate a weighted
        2D probability distribution. 

        Parameters
        ----------
        iteration : int
            Desired iteration to extract timeseries info from.

        Returns
        -------
        midpoints_x : ndarray
            Histogram midpoint bin values for target aux coordinate of dimension 0.
        midpoints_y : ndarray
            Optional histogram midpoint bin values for target aux coordinate of dimension 1.
        histogram : ndarray
            Raw histogram count values of each histogram bin. Can be later normalized as -lnP(x).
        """
        # 2D instant histogram and midpoint values for a single specified WE iteration
        X = self._get_data_array(self.Xname, self.Xindex, iteration)
        Y = self._get_data_array(self.Yname, self.Yindex, iteration)

        # 2D array to store hist counts for each timepoint in both dimensions
        histogram = np.zeros(self.bins)
        for seg in range(0, X.shape[0]):
            counts, bins_x, bins_y = np.histogram2d(X[seg], Y[seg], 
                                                    bins=self.bins, 
                                                    range=[self.histrange_x, 
                                                           self.histrange_y]
                                                    )

            if self.weighted is True:
                # multiply counts vector by weight scalar from weight array
                counts = np.multiply(counts, self.weights[iteration - self.first_iter][seg])

            # add all of the weighted walkers to total array for 
            # the resulting linear combination
            histogram = np.add(histogram, counts)

        # get bin midpoints
        midpoints_x = (bins_x[:-1] + bins_x[1:]) / 2
        midpoints_y = (bins_y[:-1] + bins_y[1:]) / 2

        # TODO: save these as instance attributes
        # this will make it easier to save into a text pdist file later

        # save midpoints and transposed histogram (corrected for plotting)
        return midpoints_x, midpoints_y, histogram.T

    def evolution_pdist(self):
        """
        Returns the pdist for 1 coordinate for the range iterations specified.

        Returns
        -------
        x, y, norm_hist : arrays
            x and y axis values, and if using Y or evolution (with only X), 
            also returns norm_hist.
            norm_hist is a 2-D matrix of the normalized histogram values.
        """
        # make array to store hist (-lnP) values for n iterations of X
        evolution_x = np.zeros((self.last_iter - self.first_iter + 1, self.bins[0]))
        positions_x = np.zeros((self.last_iter - self.first_iter + 1, self.bins[0]))

        for iter in tqdm(range(self.first_iter, self.last_iter + 1), 
                         desc="Evolution", disable=self.no_pbar):
            # account for first_iter arg for array indexing
            iter_index = iter - self.first_iter + 1
            # generate evolution x data
            center_x, counts_total_x = self.aux_to_pdist_1d(iter)
            evolution_x[iter_index - 1] = counts_total_x
            positions_x[iter_index - 1] = center_x

        # 2D evolution plot of X (Y not used if provided) per iteration        
        evolution_x = self._normalize(evolution_x)

        # bin positions along aux x, WE iteration numbers, z data
        return positions_x, np.arange(self.first_iter, self.last_iter + 1, 1), evolution_x

    # TODO: maybe don't need individual functions, maybe can handle in main
    def instant_pdist_1d(self):
        """
        Returns the x and y pdist datasets for a single iteration.

        Returns
        -------
        Xdata, y : arrays
            x (dataset) and y (pdist) axis values
        """
        center, counts_total = self.aux_to_pdist_1d(self.last_iter)
        counts_total = self._normalize(counts_total)
        return center, counts_total

    def instant_pdist_2d(self):
        """
        Returns the xyz pdist datasets for a single iteration.

        Returns
        -------
        x, y, norm_hist : arrays
            x and y axis values, and if using Y or evolution (with only X), 
            also returns norm_hist.
            norm_hist is a 2-D matrix of the normalized histogram values.
        """
        center_x, center_y, counts_total = self.aux_to_pdist_2d(self.last_iter)
        counts_total = self._normalize(counts_total)
        return center_x, center_y, counts_total

    def instant_datasets_3d(self):
        """
        Unique case where `Zname` is specified and the XYZ datasets are returned.
        For single iteration.

        Returns
        -------
        X, Y, Z : arrays 
            Raw data for each named coordinate.
        """
        X = self._get_data_array(self.Xname, self.Xindex, self.last_iter)
        # for the case where Zname is specified but not Yname
        if self.Yname is None:
            warn("`Zname` is defined but not `Yname`, using Yname=`pcoord`")
            Y = self._get_data_array("pcoord", self.Yindex, self.last_iter)
        else:
            Y = self._get_data_array(self.Yname, self.Yindex, self.last_iter)
        Z = self._get_data_array(self.Zname, self.Zindex, self.last_iter)

        return X, Y, Z

    def average_pdist_1d(self):
        """
        1 dataset: average pdist for a range of iterations.

        Returns
        -------
        x, y
            x and y axis values, and if using Y or evolution (with only X), 
            also returns norm_hist.
            norm_hist is a 2-D matrix of the normalized histogram values.
        """
        # make array to store hist (-lnP) values for n iterations of X
        evolution_x = np.zeros((self.last_iter, self.bins[0]))
        positions_x = np.zeros((self.last_iter, self.bins[0]))

        for iter in tqdm(range(self.first_iter, self.last_iter + 1), 
                         desc="Average 1D", disable=self.no_pbar):
            # generate evolution x data
            center_x, counts_total_x = self.aux_to_pdist_1d(iter)
            evolution_x[iter - 1] = counts_total_x
            positions_x[iter - 1] = center_x

        # summation of counts for all iterations : then normalize
        col_avg_x = [np.sum(col[col != np.isinf]) for col in evolution_x.T]
        col_avg_x = self._normalize(col_avg_x)

        # 1D average plot data for X
        return center_x, col_avg_x

    def average_pdist_2d(self):
        """
        2 datasets: average pdist for a range of iterations.

        Returns
        -------
        x, y, norm_hist
            x and y axis values, and if using Y or evolution (with only X), also returns norm_hist.
            norm_hist is a 2-D matrix of the normalized histogram values.
        """
        # empty array for 2D pdist
        average_xy = np.zeros(self.bins)

        # 2D avg pdist data generation
        for iter in tqdm(range(self.first_iter, self.last_iter + 1), 
                         desc="Average 2D", disable=self.no_pbar):
            center_x, center_y, counts_total_xy = self.aux_to_pdist_2d(iter)
            average_xy = np.add(average_xy, counts_total_xy)

        average_xy = self._normalize(average_xy)
        return center_x, center_y, average_xy

    def average_datasets_3d(self, interval=1):
        """
        Unique case where `Zname` is specified and the XYZ datasets are returned.
        Averaged over the iteration range.
        
        Returns
        -------
        X, Y, Z : arrays 
            Raw data for each named coordinate.
        """
        if self.Yname is None:
            warn("`Zname` is defined but not `Yname`, using Yname=`pcoord`")
            self.Yname = "pcoord"

        # arrays to be filled with values from each iteration
        # rows are for all segments, columns are each segment datapoint
        X = np.zeros((self.current_particles, self.tau))
        Y = np.zeros((self.current_particles, self.tau))
        Z = np.zeros((self.current_particles, self.tau))

        # loop each iteration
        seg_start = 0
        for iter in tqdm(range(self.first_iter, self.last_iter + 1), 
                         desc="Average 3D", disable=self.no_pbar):
            # then go through and add all segments/walkers in the iteration
            X[seg_start:seg_start + self.n_particles[iter - 1]] = \
                self._get_data_array(self.Xname, self.Xindex, iter)
            Y[seg_start:seg_start + self.n_particles[iter - 1]] = \
                self._get_data_array(self.Yname, self.Yindex, iter)
            Z[seg_start:seg_start + self.n_particles[iter - 1]] = \
                self._get_data_array(self.Zname, self.Zindex, iter)

            # keeps track of position in the seg_total length based arrays
            seg_start += self.n_particles[iter - 1]

        # 3D average datasets using all available data (can more managable with interval)
        return X[::interval], Y[::interval], Z[::interval]

    def get_all_weights(self):
        """
        Returns an 1D array of the weight for every frame of each tau 
        for all segments of all iterations specified.

        Returns
        -------
        weights_expanded : array
        """
        # weights per seg of each iter, but need for each frame
        weights_1d = np.concatenate(self.weights)

        # need each weight value to be repeated for each tau (e.g. 100 + 1) 
        # will be same shape as X or Y made into 1d shape
        weights_expanded = np.zeros(self.tau * self.current_particles)

        # loop over all ps intervals up to tau in each segment
        weight_index = 0
        for seg in weights_1d:
            # TODO: can I do this without the unused loop?
            for frame in range(self.tau):
                weights_expanded[weight_index] = seg
                weight_index += 1

        return weights_expanded

    # TODO: option for data and weight output for a single iteration (iteration=None)
    # wait, isn't that already available in _get_data_array?
    def get_total_data_array(self, name, index=0, interval=1, reshape=True):
        """
        Loop through all iterations specified and get a 1d raw data array.
        # TODO: this could be organized better with my other methods
        maybe I can separate the helper functions into another class
        for extracting and moving data around, this pdist class could
        be used strictly for making pdists from a nice and standard data
        array input that is handled by the H5_Processing class

        Parameters
        ----------
        name : str
            Name of data from h5 file such as `pcoord` or an aux dataset.
        index : int
            Index of the data from h5 file.
        interval : int
            If more sparse data is needed for efficiency.
        reshape : bool
            Option to reshape into 1d array instead of each seg for all tau values.

        Returns
        -------
        data : 1d array
            Raw (unweighted) data array for the name specified.
        """
        data = np.zeros((self.current_particles, self.tau))
        # account for non-pcoord input strings
        if name != "pcoord":
            name = "auxdata/" + name
    
        # loop each iteration
        seg_start = 0
        for iter in tqdm(range(self.first_iter, self.last_iter + 1), 
                         desc="Getting Data Array", disable=self.no_pbar):
            # then go through and add all segments/walkers in the iteration
            data[seg_start : seg_start + self.n_particles[iter - 1]] = \
                self._get_data_array(name, index, iter)
            
            # keeps track of position in the seg_total length based arrays
            seg_start += self.n_particles[iter - 1]

        if reshape:
            return data[::interval].reshape(-1,1)
        else:
            return data[::interval]

    def reshape_total_data_array(self, array):
        """
        Take an input 1d array of the data values at every segment for each
        iteration, and reshape them to make pdists.

        Parameters
        ----------
        array : 1d array
            Data values at every segment for each iteration.

        Returns
        -------
        array : ndarray
            Now rows = segments, columns = frame until tau, depth = data dimensions.
        """
        # try except block for input from data array
        # which can be the correct shape if pulled from westpa (e.g. 100 ps + 1)
        # or if from agg MD sim, will just be (e.g. 100 ps)
        # also adding -1 in z dim for extra depth dimension compatibility

        # TODO: change total particles to iteration range to be able to use iter args with data arrays

        try:
            array = array.reshape(self.total_particles, self.tau, -1)
        # e.g. ValueError: cannot reshape array of size 303000 into shape (3000,100,newaxis)
        except ValueError as e:
            array = array.reshape(self.total_particles, self.tau - 1, -1)
            message = "\nYou may be using an input data array which did not include the rst file datapoints. " + \
                      "\nThis may be fine, but note that you shouldn't create a new H5 file using this array."
            warn(e + message)
            # TODO: does this work?
            # the case where the array does not have rst data included
            # put the new first column as the first value of each row (segment)
            # TODO: this is a temp hack for the no rst shape data
            # noting that both arrays must have same ndims for hstack
            #print(f"original shape: {data.shape}")
            #print(f"to stack shape: {data[:,0,:]}")
            array = np.hstack((np.atleast_3d(array[:,0,:]), array)) # TODO: test this
            #print(f"new shape: {data.shape}")

        # TODO: the above works to solve the shape issue but if I wanted to fill out a new dataset in
        # the h5 file, it would be missing the first value, which links walkers.
        # maybe I can use the parent IDs to link it manually, but note I would have to
        # go through and parse by my self.n_particles array to separate iterations.
        # put conditional if shape[1] = tau vs tau - 1 for creating dataset (to add parent data point)
        # note that the first iteration I need to pull from somewhere else? It's calculated from the
        # original bstate file

        # Note, if the user includes the rst files like WESTPA does, it should look and process fine

        return array
        
    def pdist(self):
        """
        Main public method with pdist generation controls.
        """ 
        # option to zero weight out specific basis states
        if self.skip_basis is not None:
            self.n_bstates = self.f["ibstates/index"]["n_bstates"]
            try: 
                self.weights = self._new_weights_from_skip_basis()
            # if the wrong amount of args are input and != n_bstates
            except IndexError as e:
                message = f"IndexError ({e}) for bstate input ({self.skip_basis}): " + \
                          f"Did you use the correct amount of bstates {self.n_bstates}?"
                warn(message)

        # reshape 1d input raw data array (if given) into 3d array
        if isinstance(self.Xname, np.ndarray):
            self.Xname = self.reshape_total_data_array(self.Xname)
        if isinstance(self.Yname, np.ndarray):
            self.Yname = self.reshape_total_data_array(self.Yname)
        if isinstance(self.Zname, np.ndarray):
            self.Zname = self.reshape_total_data_array(self.Zname)

        # TODO: could make this it's own method
        # if requested, save out a new H5 file with the input data array in new aux name
        if self.H5save_out is not None:
            for iter in range(self.first_iter, self.last_iter + 1):
                if self.Xsave_name:
                    self._get_data_array(self.Xname, self.Xindex, iter, self.H5save_out, self.Xsave_name)
                if self.Ysave_name:
                    self._get_data_array(self.Yname, self.Yindex, iter, self.H5save_out, self.Ysave_name)
                if self.Zsave_name:
                    self._get_data_array(self.Zname, self.Zindex, iter, self.H5save_out, self.Zsave_name)

        # TODO: need to consolidate the Y 2d vs 1d stuff somehow

        # TODO: if I can get rid of this or optimize it, I can then use the 
            # original methods of each pdist by themselves
        # TODO: only if histrange is None
        if self.histrange_x is None:
            # get the optimal histrange
            self.histrange_x = self._get_histrange(self.Xname, self.Xindex)
        # if using 2D pdist
        # TODO: needs to handle array input or None input
        if isinstance(self.Yname, (str, np.ndarray)) and self.histrange_y is None:
            self.histrange_y = self._get_histrange(self.Yname, self.Yindex)

        # TODO: need a better way to always return XYZ (currently using ones)
        #       this is needed for easy testing with uniform XYZ 3 array returns
        if self.data_type == "evolution":
            return self.evolution_pdist()
        elif self.data_type == "instant":
            if self.Yname and self.Zname:
                return self.instant_datasets_3d()
            elif self.Yname:
                return self.instant_pdist_2d()
            else:
                X, Y = self.instant_pdist_1d()
                return X, Y, np.ones((self.first_iter, self.last_iter))
        elif self.data_type == "average":
            # attemts to say, if not None, but has to be compatible with str and arrays
            if isinstance(self.Yname, (str, np.ndarray)) and isinstance(self.Zname, (str, np.ndarray)):
                return self.average_datasets_3d()
            elif isinstance(self.Yname, (str, np.ndarray)):
                return self.average_pdist_2d()
            else:
                X, Y = self.average_pdist_1d()
                return X, Y, np.ones((self.first_iter, self.last_iter))
            
        # safely close h5 file
        self.h5.close()

#if __name__ == "__main__":
    # total_array_out = np.loadtxt("p53_X_array.txt")
    # original_array = np.loadtxt("p53_X_array_noreshape.txt")
    
    # h5 = H5_Pdist("data/p53.h5", data_type="evolution")
    # TODO: test Zname with data_array


"""
Plot MD pdists.

TODO: specify option for:
    timeseries (option for KDE side plot and option for stdev vs all reps)
    pdist (1D hist, 1D KDE, + others from H5_Plot)
    others?
"""

from .md_pdist import MD_Pdist

# relative import of wedap H5_Plot class (TODO: alt import?)
from wedap import H5_Plot

# subclass H5_Plot to have access to those pdist plots
# also multiple inheritance for features from MD_Pdist as well
class MD_Plot(H5_Plot, MD_Pdist):
    """
    These methods provide various plotting options for pdist data.
    """
    def __init__(self, *args, **kwargs):

        # for jointplots, save original p_units and run pdist with raw
        # only if jointplot var exists and is True
        if "jointplot" in kwargs:
            # separate since if it doesn't exist can't index to check if True
            if kwargs["jointplot"] is True:
                # with user specified p_units, save it
                if "p_units" in kwargs:
                    og_p_units = kwargs["p_units"]
                # always set p_units to 'raw' for joint plots
                kwargs["p_units"] = "raw"

        # initialize md pdist
        MD_Pdist.__init__(self, *args, **kwargs)
        # generate xyz arrays
        self.X, self.Y, self.Z = self.pdist()

        # change back to original p_units
        # only if jointplot var exists and is True
        if "jointplot" in kwargs:
            # separate since if it doesn't exist can't index to check if True
            if kwargs["jointplot"] is True:
                # if the original p_units from user were provided and saved
                if "p_units" in kwargs and "og_p_units" in locals():
                    kwargs["p_units"] = og_p_units
                    self.p_units = og_p_units
                # otherwise default to kT
                else:
                    kwargs["p_units"] = "kT"
                    self.p_units = "kT"

        # run H5_Plot initialization
        H5_Plot.__init__(self, self.X, self.Y, self.Z, *args, **kwargs)